"""
A :class:`Group` can contain sub-:class:`Group`\\s and/or :class:`~msl.io.dataset.Dataset`\\s.
"""
import re

from .dataset import Dataset
from .dataset_logging import DatasetLogging
from .vertex import Vertex


class Group(Vertex):

    def __init__(self, name, parent, read_only, **metadata):
        """A :class:`Group` can contain sub-:class:`Group`\\s and/or :class:`~msl.io.dataset.Dataset`\\s.

        Do not instantiate directly. Create a new :class:`Group` using
        :meth:`~msl.io.group.Group.create_group`.

        Parameters
        ----------
        name : :class:`str`
            The name of this :class:`Group`. Uses a naming convention analogous to UNIX
            file systems where each :class:`~msl.io.group.Group` can be thought
            of as a directory and where every subdirectory is separated from its
            parent directory by the ``'/'`` character.
        parent : :class:`Group`
            The parent :class:`Group` to this :class:`Group`.
        read_only : :class:`bool`
            Whether the :class:`Group` is to be accessed in read-only mode.
        **metadata
            Key-value pairs that are used to create the :class:`~msl.io.metadata.Metadata`
            for this :class:`Group`.
        """
        super(Group, self).__init__(name, parent, read_only, **metadata)

    def __repr__(self):
        g = len(list(self.groups()))
        d = len(list(self.datasets()))
        m = len(self.metadata)
        return '<Group {!r} ({} groups, {} datasets, {} metadata)>'.format(self._name, g, d, m)

    def __getitem__(self, item):
        if item and not item[0] == '/':
            item = '/' + item
        try:
            return self._mapping[item]
        except KeyError:
            pass  # raise a more detailed error message below
        self._raise_key_error(item)

    def __getattr__(self, item):
        try:
            return self.__getitem__('/' + item)
        except KeyError as e:
            msg = str(e)
        raise AttributeError(msg)

    def __delattr__(self, item):
        try:
            return self.__delitem__('/' + item)
        except KeyError as e:
            msg = str(e)
        raise AttributeError(msg)

    @staticmethod
    def is_dataset(obj):
        """Test whether an object is a :class:`~msl.io.dataset.Dataset`.

        Parameters
        ----------
        obj : :class:`object`
            The object to test.

        Returns
        -------
        :class:`bool`
            Whether `obj` is an instance of :class:`~msl.io.dataset.Dataset`.
        """
        return isinstance(obj, Dataset)

    @staticmethod
    def is_dataset_logging(obj):
        """Test whether an object is a :class:`~msl.io.dataset_logging.DatasetLogging`.

        Parameters
        ----------
        obj : :class:`object`
            The object to test.

        Returns
        -------
        :class:`bool`
            Whether `obj` is an instance of :class:`~msl.io.dataset_logging.DatasetLogging`.
        """
        return isinstance(obj, DatasetLogging)

    @staticmethod
    def is_group(obj):
        """Test whether an object is a :class:`~msl.io.group.Group`.

        Parameters
        ----------
        obj : :class:`object`
            The object to test.

        Returns
        -------
        :class:`bool`
            Whether `obj` is an instance of :class:`~msl.io.group.Group`.
        """
        return isinstance(obj, Group)

    def datasets(self, exclude=None, include=None, flags=0):
        """Get the :class:`~msl.io.dataset.Dataset`\\s in this :class:`Group`.

        Parameters
        ----------
        exclude : :class:`str`, optional
            A regex pattern to use to exclude :class:`~msl.io.dataset.Dataset`\\s.
            The :func:`re.search` function is used to compare the `exclude` regex
            pattern with the `name` of each :class:`~msl.io.dataset.Dataset`. If
            there is a match, the :class:`~msl.io.dataset.Dataset` is not yielded.
        include : :class:`str`, optional
            A regex pattern to use to include :class:`~msl.io.dataset.Dataset`\\s.
            The :func:`re.search` function is used to compare the `include` regex
            pattern with the `name` of each :class:`~msl.io.dataset.Dataset`. If
            there is a match, the :class:`~msl.io.dataset.Dataset` is yielded.
        flags : :class:`int`, optional
            Regex flags that are passed to :func:`re.compile`.

        Yields
        ------
        :class:`~msl.io.dataset.Dataset`
            The filtered :class:`~msl.io.dataset.Dataset`\\s based on the
            `exclude` and `include` regex patterns. The `exclude` pattern
            has more precedence than the `include` pattern if there is a
            conflict.
        """
        e = False if exclude is None else re.compile(exclude, flags=flags)
        i = False if include is None else re.compile(include, flags=flags)
        for obj in self._mapping.values():
            if self.is_dataset(obj):
                if e and e.search(obj.name):
                    continue
                if i and not i.search(obj.name):
                    continue
                yield obj

    def groups(self, exclude=None, include=None, flags=0):
        """Get the sub-:class:`.Group`\\s of this :class:`.Group`.

        Parameters
        ----------
        exclude : :class:`str`, optional
            A regex pattern to use to exclude :class:`.Group`\\s. The
            :func:`re.search` function is used to compare the `exclude` regex
            pattern with the `name` of each :class:`.Group`. If there is a match,
            the :class:`.Group` is not yielded.
        include : :class:`str`, optional
            A regex pattern to use to include :class:`.Group`\\s. The
            :func:`re.search` function is used to compare the `include` regex
            pattern with the `name` of each :class:`.Group`. If there is a match,
            the :class:`.Group` is yielded.
        flags : :class:`int`, optional
            Regex flags that are passed to :func:`re.compile`.

        Yields
        ------
        :class:`Group`
            The filtered :class:`.Group`\\s based on the `exclude` and `include`
            regex patterns. The `exclude` pattern has more precedence than the
            `include` pattern if there is a conflict.
        """
        e = False if exclude is None else re.compile(exclude, flags=flags)
        i = False if include is None else re.compile(include, flags=flags)
        for obj in self._mapping.values():
            if self.is_group(obj):
                if e and e.search(obj.name):
                    continue
                if i and not i.search(obj.name):
                    continue
                yield obj

    def descendants(self):
        """Get all descendant (children) :class:`.Group`\\s of this :class:`.Group`.

        Yields
        ------
        :class:`.Group`
            The descendants of this :class:`.Group`.
        """
        for obj in self._mapping.values():
            if self.is_group(obj):
                yield obj

    def ancestors(self):
        """Get all ancestor (parent) :class:`.Group`\\s of this :class:`.Group`.

        Yields
        ------
        :class:`.Group`
            The ancestors of this :class:`.Group`.
        """
        parent = self.parent
        while parent is not None:
            yield parent
            parent = parent.parent

    def add_group(self, name, group):
        """Add a :class:`Group`.

        Automatically creates the ancestor :class:`Group`\\s if they do not exist.

        Parameters
        ----------
        name : :class:`str`
            The name of the new :class:`Group` to add.
        group : :class:`Group`
            The :class:`Group` to add. The :class:`~msl.io.dataset.Dataset`\\s and
            :class:`~msl.io.metadata.Metadata` that are contained within the
            `group` will be copied.
        """
        if not isinstance(group, Group):
            raise TypeError('Must pass in a Group object, got {!r}'.format(group))

        name = '/' + name.strip('/')

        if not group:  # no sub-Groups or Datasets, only add the Metadata
            self.create_group(name + group.name, **group.metadata.copy())
            return

        for key, vertex in group.items():
            n = name + key
            if self.is_group(vertex):
                self.create_group(n, read_only=vertex.read_only, **vertex.metadata.copy())
            else:  # must be a Dataset
                self.create_dataset(
                    n, read_only=vertex.read_only, data=vertex.data.copy(), **vertex.metadata.copy()
                )

    def create_group(self, name, read_only=None, **metadata):
        """Create a new :class:`Group`.

        Automatically creates the ancestor :class:`Group`\\s if they do not exist.

        Parameters
        ----------
        name : :class:`str`
            The name of the new :class:`Group`.
        read_only : :class:`bool`, optional
            Whether to create this :class:`Group` in read-only mode.
            If :data:`None` then uses the mode for this :class:`Group`.
        **metadata
            Key-value pairs that are used to create the :class:`~msl.io.metadata.Metadata`
            for this :class:`Group`.

        Returns
        -------
        :class:`Group`
            The new :class:`Group` that was created.
        """
        read_only, metadata = self._check(read_only, **metadata)
        name, parent = self._create_ancestors(name, read_only)
        return Group(name, parent, read_only, **metadata)

    def require_group(self, name, read_only=None, **metadata):
        """Require that a :class:`Group` exists.

        If the :class:`Group` exists then it will be returned if it does not exist
        then it is created.

        Automatically creates the ancestor :class:`Group`\\s if they do not exist.

        Parameters
        ----------
        name : :class:`str`
            The name of the :class:`Group`.
        read_only : :class:`bool`, optional
            Whether to return the :class:`Group` in read-only mode.
            If :data:`None` then uses the mode for this :class:`Group`.
        **metadata
            Key-value pairs that are used as :class:`~msl.io.metadata.Metadata`
            for this :class:`Group`.

        Returns
        -------
        :class:`Group`
            The :class:`Group` that was created or that already existed.
        """
        name = '/' + name.strip('/')
        group_name = name if self.parent is None else self.name + name
        for group in self.groups():
            if group.name == group_name:
                if read_only is not None:
                    group.read_only = read_only
                group.add_metadata(**metadata)
                return group
        return self.create_group(name, read_only=read_only, **metadata)

    def add_dataset(self, name, dataset):
        """Add a :class:`~msl.io.dataset.Dataset`.

        Automatically creates the ancestor :class:`Group`\\s if they do not exist.

        Parameters
        ----------
        name : :class:`str`
            The name of the new :class:`~msl.io.dataset.Dataset` to add.
        dataset : :class:`~msl.io.dataset.Dataset`
            The :class:`~msl.io.dataset.Dataset` to add. The :class:`~msl.io.dataset.Dataset`
            and the :class:`~msl.io.metadata.Metadata` are copied.
        """
        if not isinstance(dataset, Dataset):
            raise TypeError('Must pass in a Dataset object, got {!r}'.format(dataset))

        name = '/' + name.strip('/')
        self.create_dataset(
            name, read_only=dataset.read_only, data=dataset.data.copy(), **dataset.metadata.copy()
        )

    def create_dataset(self, name, read_only=None, **kwargs):
        """Create a new :class:`~msl.io.dataset.Dataset`.

        Automatically creates the ancestor :class:`Group`\\s if they do not exist.

        Parameters
        ----------
        name : :class:`str`
            The name of the new :class:`~msl.io.dataset.Dataset`.
        read_only : :class:`bool`, optional
            Whether to create this :class:`~msl.io.dataset.Dataset` in read-only mode.
            If :data:`None` then uses the mode for this :class:`Group`.
        **kwargs
            Key-value pairs that are passed to :class:`~msl.io.dataset.Dataset`.

        Returns
        -------
        :class:`~msl.io.dataset.Dataset`
            The new :class:`~msl.io.dataset.Dataset` that was created.
        """
        read_only, kwargs = self._check(read_only, **kwargs)
        name, parent = self._create_ancestors(name, read_only)
        return Dataset(name, parent, read_only, **kwargs)

    def require_dataset(self, name, read_only=None, **kwargs):
        """Require that a :class:`~msl.io.dataset.Dataset` exists.

        If the :class:`~msl.io.dataset.Dataset` exists then it will be returned
        if it does not exist then it is created.

        Automatically creates the ancestor :class:`Group`\\s if they do not exist.

        Parameters
        ----------
        name : :class:`str`
            The name of the :class:`~msl.io.dataset.Dataset`.
        read_only : :class:`bool`, optional
            Whether to create this :class:`~msl.io.dataset.Dataset` in read-only mode.
            If :data:`None` then uses the mode for this :class:`Group`.
        **kwargs
            Key-value pairs that are passed to :class:`~msl.io.dataset.Dataset`.

        Returns
        -------
        :class:`~msl.io.dataset.Dataset`
            The :class:`~msl.io.dataset.Dataset` that was created or that already existed.
        """
        name = '/' + name.strip('/')
        dataset_name = name if self.parent is None else self.name + name
        for dataset in self.datasets():
            if dataset.name == dataset_name:
                if read_only is not None:
                    dataset.read_only = read_only
                if kwargs:  # only add the kwargs that should be Metadata
                    for kw in ['shape', 'dtype', 'buffer', 'offset', 'strides', 'order', 'data']:
                        kwargs.pop(kw, None)
                dataset.add_metadata(**kwargs)
                return dataset
        return self.create_dataset(name, read_only=read_only, **kwargs)

    def add_dataset_logging(self, name, dataset_logging):
        """Add a :class:`~msl.io.dataset_logging.DatasetLogging`.

        Automatically creates the ancestor :class:`Group`\\s if they do not exist.

        Parameters
        ----------
        name : :class:`str`
            The name of the new :class:`~msl.io.dataset_logging.DatasetLogging` to add.
        dataset_logging : :class:`~msl.io.dataset_logging.DatasetLogging`
            The :class:`~msl.io.dataset_logging.DatasetLogging` to add. The
            :class:`~msl.io.dataset_logging.DatasetLogging` and the
            :class:`~msl.io.metadata.Metadata` are copied.
        """
        if not isinstance(dataset_logging, DatasetLogging):
            raise TypeError('Must pass in a DatasetLogging object, got {!r}'.format(dataset_logging))

        name = '/' + name.strip('/')
        self.create_dataset_logging(
            name,
            level=dataset_logging.level,
            attributes=dataset_logging.attributes,
            logger=dataset_logging.logger,
            date_fmt=dataset_logging.date_fmt,
            data=dataset_logging.data.copy(),
            **dataset_logging.metadata.copy()
        )

    def create_dataset_logging(self, name, level='INFO', attributes=None, logger=None, date_fmt=None, **kwargs):
        """Create a :class:`~msl.io.dataset.Dataset` that handles :mod:`logging` records.

        Automatically creates the ancestor :class:`Group`\\s if they do not exist.

        Parameters
        ----------
        name : :class:`str`
            A name to associate with the :class:`~msl.io.dataset.Dataset`.
        level : :class:`int` or :class:`str`, optional
            The :ref:`logging level <levels>` to use.
        attributes : :class:`list` or :class:`tuple` of :class:`str`, optional
            The :ref:`attribute names <logrecord-attributes>` to include in the
            :class:`~msl.io.dataset.Dataset` for each :ref:`logging record <log-record>`.
            If :data:`None` then uses ``asctime``, ``levelname``, ``name``, and ``message``.
        logger : :class:`~logging.Logger`, optional
            The :class:`~logging.Logger` that the :class:`~msl.io.dataset_logging.DatasetLogging` object
            will be added to. If :data:`None` then it is added to the ``root`` :class:`~logging.Logger`.
        date_fmt : :class:`str`, optional
            The :class:`~datetime.datetime` :ref:`format code <strftime-strptime-behavior>`
            to use to represent the ``asctime`` :ref:`attribute <logrecord-attributes>` in.
            If :data:`None` then uses the ISO 8601 format ``'%Y-%m-%dT%H:%M:%S.%f'``.
        **kwargs
            Additional keyword arguments are passed to :class:`~msl.io.dataset.Dataset`.
            The default behaviour is to append every :ref:`logging record <log-record>`
            to the :class:`~msl.io.dataset.Dataset`. This guarantees that the size of the
            :class:`~msl.io.dataset.Dataset` is equal to the number of
            :ref:`logging records <log-record>` that were added to it. However, this behaviour
            can decrease the performance if many :ref:`logging records <log-record>` are
            added often because a copy of the data in the :class:`~msl.io.dataset.Dataset` is
            created for each :ref:`logging record <log-record>` that is added. You can improve
            the performance by specifying an initial size of the :class:`~msl.io.dataset.Dataset`
            by including a `shape` or a `size` keyword argument. This will also automatically
            create additional empty rows in the :class:`~msl.io.dataset.Dataset`, that is
            proportional to the size of the :class:`~msl.io.dataset.Dataset`, if the size of the
            :class:`~msl.io.dataset.Dataset` needs to be increased. If you do this then you will
            want to call :meth:`~msl.io.dataset_logging.DatasetLogging.remove_empty_rows` before
            writing :class:`~msl.io.dataset_logging.DatasetLogging` to a file or interacting
            with the data in :class:`~msl.io.dataset_logging.DatasetLogging` to remove the
            extra rows that were created.

        Returns
        -------
        :class:`~msl.io.dataset_logging.DatasetLogging`
            The :class:`~msl.io.dataset_logging.DatasetLogging` that was created.

        Examples
        --------
        >>> import logging
        >>> from msl.io import JSONWriter
        >>> logger = logging.getLogger('my_logger')
        >>> root = JSONWriter()
        >>> log_dset = root.create_dataset_logging('log')
        >>> logger.info('hi')
        >>> logger.error('cannot do that!')
        >>> log_dset.data
        array([(..., 'INFO', 'my_logger', 'hi'), (..., 'ERROR', 'my_logger', 'cannot do that!')],
              dtype=[('asctime', 'O'), ('levelname', 'O'), ('name', 'O'), ('message', 'O')])

        Get all ``ERROR`` :ref:`logging records <log-record>`

        >>> errors = log_dset[log_dset['levelname'] == 'ERROR']
        >>> print(errors)
        [(..., 'ERROR', 'my_logger', 'cannot do that!')]

        Stop the :class:`~msl.io.dataset_logging.DatasetLogging` object
        from receiving :ref:`logging records <log-record>`

        >>> log_dset.remove_handler()
        """
        read_only, metadata = self._check(False, **kwargs)
        name, parent = self._create_ancestors(name, read_only)
        if attributes is None:
            # if the default attribute names are changed then update the `attributes`
            # description in the docstring of create_dataset_logging() and require_dataset_logging()
            attributes = ['asctime', 'levelname', 'name', 'message']
        if date_fmt is None:
            # if the default date_fmt is changed then update the `date_fmt`
            # description in the docstring of create_dataset_logging() and require_dataset_logging()
            date_fmt = '%Y-%m-%dT%H:%M:%S.%f'
        return DatasetLogging(name, parent, level=level, attributes=attributes,
                              logger=logger, date_fmt=date_fmt, **metadata)

    def require_dataset_logging(self, name, level='INFO', attributes=None, logger=None, date_fmt=None, **kwargs):
        """Require that a :class:`~msl.io.dataset.Dataset` exists for handling :mod:`logging` records.

        If the :class:`~msl.io.dataset_logging.DatasetLogging` exists then it will be returned
        if it does not exist then it is created.

        Automatically creates the ancestor :class:`Group`\\s if they do not exist.

        Parameters
        ----------
        name : :class:`str`
            A name to associate with the :class:`~msl.io.dataset.Dataset`.
        level : :class:`int` or :class:`str`, optional
            The :ref:`logging level <levels>` to use.
        attributes : :class:`list` or :class:`tuple` of :class:`str`, optional
            The :ref:`attribute names <logrecord-attributes>` to include in the
            :class:`~msl.io.dataset.Dataset` for each :ref:`logging record <log-record>`.
            If the :class:`~msl.io.dataset.Dataset` exists and if `attributes`
            are specified, and they do not match those of the existing
            :class:`~msl.io.dataset.Dataset`, then a :exc:`ValueError` is raised.
            If :data:`None` and the :class:`~msl.io.dataset.Dataset` does not exist
            then uses ``asctime``, ``levelname``, ``name``, and ``message``.
        logger : :class:`~logging.Logger`, optional
            The :class:`~logging.Logger` that the :class:`~msl.io.dataset_logging.DatasetLogging` object
            will be added to. If :data:`None` then it is added to the ``root`` :class:`~logging.Logger`.
        date_fmt : :class:`str`, optional
            The :class:`~datetime.datetime` :ref:`format code <strftime-strptime-behavior>`
            to use to represent the ``asctime`` :ref:`attribute <logrecord-attributes>` in.
            If :data:`None` then uses the ISO 8601 format ``'%Y-%m-%dT%H:%M:%S.%f'``.
        **kwargs
            Additional keyword arguments are passed to :class:`~msl.io.dataset.Dataset`.
            The default behaviour is to append every :ref:`logging record <log-record>`
            to the :class:`~msl.io.dataset.Dataset`. This guarantees that the size of the
            :class:`~msl.io.dataset.Dataset` is equal to the number of
            :ref:`logging records <log-record>` that were added to it. However, this behaviour
            can decrease the performance if many :ref:`logging records <log-record>` are
            added often because a copy of the data in the :class:`~msl.io.dataset.Dataset` is
            created for each :ref:`logging record <log-record>` that is added. You can improve
            the performance by specifying an initial size of the :class:`~msl.io.dataset.Dataset`
            by including a `shape` or a `size` keyword argument. This will also automatically
            create additional empty rows in the :class:`~msl.io.dataset.Dataset`, that is
            proportional to the size of the :class:`~msl.io.dataset.Dataset`, if the size of the
            :class:`~msl.io.dataset.Dataset` needs to be increased. If you do this then you will
            want to call :meth:`~msl.io.dataset_logging.DatasetLogging.remove_empty_rows` before
            writing :class:`~msl.io.dataset_logging.DatasetLogging` to a file or interacting
            with the data in :class:`~msl.io.dataset_logging.DatasetLogging` to remove the
            extra rows that were created.

        Returns
        -------
        :class:`~msl.io.dataset_logging.DatasetLogging`
            The :class:`~msl.io.dataset_logging.DatasetLogging` that was created or
            that already existed.
        """
        name = '/' + name.strip('/')
        dataset_name = name if self.parent is None else self.name + name
        for dataset in self.datasets():
            if dataset.name == dataset_name:
                if ('logging_level' not in dataset.metadata) or \
                        ('logging_level_name' not in dataset.metadata) or \
                        ('logging_date_format' not in dataset.metadata):
                    raise ValueError('The required Dataset was found but it is not used for logging')

                if attributes and (dataset.dtype.names != tuple(attributes)):
                    raise ValueError('The attribute names of the existing '
                                     'logging Dataset are {} which does not equal {}'
                                     .format(dataset.dtype.names, tuple(attributes)))

                if isinstance(dataset, DatasetLogging):
                    return dataset

                # replace the existing Dataset with a new DatasetLogging object
                meta = dataset.metadata.copy()
                data = dataset.data.copy()

                # remove the existing Dataset from its descendants, itself and its ancestors
                groups = tuple(self.descendants()) + (self,) + tuple(self.ancestors())
                for group in groups:
                    for dset in group.datasets():
                        if dset is dataset:
                            key = '/' + dset.name.lstrip(group.name)
                            del group._mapping[key]

                # temporarily make this Group not in read-only mode
                original_read_only_mode = bool(self._read_only)
                self._read_only = False
                kwargs.update(meta)
                dset = self.create_dataset_logging(name, level=level, attributes=data.dtype.names,
                                                   logger=logger, date_fmt=meta.logging_date_format,
                                                   data=data, **kwargs)
                self._read_only = original_read_only_mode
                return dset

        return self.create_dataset_logging(name, level=level, attributes=attributes,
                                           logger=logger, date_fmt=date_fmt, **kwargs)

    def remove(self, name):
        """Remove a :class:`Group` or a :class:`~msl.io.dataset.Dataset`.

        Parameters
        ----------
        name : :class:`str`
            The name of the :class:`Group` or :class:`~msl.io.dataset.Dataset` to remove.

        Returns
        -------
        :class:`Group`, :class:`~msl.io.dataset.Dataset` or :data:`None`
            The :class:`Group` or :class:`~msl.io.dataset.Dataset` that was
            removed or :data:`None` if there was no :class:`Group` or
            :class:`~msl.io.dataset.Dataset` with the specified `name`.
        """
        name = '/' + name.strip('/')
        return self.pop(name, None)

    def _check(self, read_only, **kwargs):
        self._raise_if_read_only()
        kwargs.pop('parent', None)
        if read_only is None:
            return self._read_only, kwargs
        return read_only, kwargs

    def _create_ancestors(self, name, read_only):
        # automatically create the ancestor Groups if they do not already exist
        names = name.strip('/').split('/')
        parent = self
        for n in names[:-1]:
            if n not in parent:
                parent = Group(n, parent, read_only)
            else:
                parent = parent[n]
        return names[-1], parent



"""
A vertex in a tree_.

.. _tree: https://en.wikipedia.org/wiki/Tree_(graph_theory)
"""
import os

from .dictionary import Dictionary
from .metadata import Metadata


class Vertex(Dictionary):

    def __init__(self, name, parent, read_only, **metadata):
        """A vertex in a tree_.

        Parameters
        ----------
        name : :class:`str`
            The name of this vertex.
        parent : :class:`~msl.io.group.Group`
            The parent of this vertex.
        read_only : :class:`bool`
            Whether this vertex is in read-only mode.
        **metadata
            Key-value pairs that are used to create the :class:`~msl.io.metadata.Metadata`
            for this :class:`~msl.io.vertex.Vertex`.
        """
        super(Vertex, self).__init__(read_only)

        if not name:
            raise ValueError('The vertex name must be a non-empty string')

        if parent is not None:
            if '/' in name:
                raise ValueError('The vertex name cannot contain the "/" character')

            # use a path name similar to a UNIX file system
            if parent.name.endswith('/'):
                name = parent.name + name
            else:
                name = parent.name + '/' + name

            # notify all ancestors that this vertex was created
            i = 0
            ancestor = parent
            name_split = name.split('/')
            while ancestor is not None:
                i += 1
                key = '/' + '/'.join(name_split[-i:])
                if key in ancestor._mapping:
                    raise ValueError('The name of this vertex, {!r}, is not unique'.format(key))
                ancestor._mapping[key] = self
                ancestor = ancestor._parent

        self._name = name
        self._parent = parent
        self._metadata = Metadata(read_only, name, **metadata)

    def __delitem__(self, item):
        self._raise_if_read_only()
        if item and not item[0] == '/':
            item = '/' + item

        try:
            popped = self._mapping.pop(item)
        except KeyError:
            pass  # raise a more detailed error message below
        else:

            # use recursion to delete the reference to
            # `popped` from the head of this Vertex
            head, tail = os.path.split(item)
            if head != '/':
                assert self[head].pop(tail) is popped

            def notify_ancestors(obj):
                # delete all references to `obj` from the
                # ancestors of this Vertex
                ancestor = self._parent
                while ancestor is not None:
                    for k, v in list(ancestor.items()):
                        if obj is v:
                            del ancestor._mapping[k]
                    ancestor = ancestor._parent

            notify_ancestors(popped)

            # delete all descendant of this Vertex
            # (this is necessary if the popped item is a Group)
            for name, vertex in list(self.items()):
                if vertex.name.startswith(popped.name):
                    vertex = self._mapping.pop(name)
                    notify_ancestors(vertex)

            return

        self._raise_key_error(item)

    @property
    def read_only(self):
        """:class:`bool`: Whether this :class:`~msl.io.vertex.Vertex` is in read-only mode.

        Setting this value will also update all sub-:class:`~msl.io.group.Group`\\s
        and sub-:class:`~msl.io.dataset.Dataset`\\s to be in the same mode.
        """
        return self._read_only

    @read_only.setter
    def read_only(self, value):
        val = bool(value)

        self._read_only = val
        self._metadata.read_only = val

        # update all descendants of this vertex
        for obj in self._mapping.values():
            obj.read_only = val

    @property
    def name(self):
        """:class:`str`: The name of this :class:`~msl.io.vertex.Vertex`."""
        return self._name

    @property
    def parent(self):
        """:class:`~msl.io.group.Group`: The parent of this :class:`~msl.io.vertex.Vertex`."""
        return self._parent

    @property
    def metadata(self):
        """:class:`~msl.io.metadata.Metadata`: The metadata associated with this :class:`~msl.io.vertex.Vertex`."""
        return self._metadata

    def add_metadata(self, **metadata):
        """Add key-value pairs to the :class:`~msl.io.metadata.Metadata` for this :class:`~msl.io.vertex.Vertex`."""
        self._metadata.update(**metadata)



"""
JSON encoder that writes a 1-dimensional list on a single line.

Must have a Python 2 and Python 3 version of this file to avoid syntax errors.

The Python 3 version uses "yield from".

This is the Python 3 version copied from the builtin json.encoder._make_iterencode function
and modified slightly for a 1-dimensional list.
"""


def _make_iterencode(markers, _default, _encoder, _indent, _floatstr,
        _key_separator, _item_separator, _sort_keys, _skipkeys, _one_shot,
        ## HACK: hand-optimized bytecode; turn globals into locals
        ValueError=ValueError,
        dict=dict,
        float=float,
        id=id,
        int=int,
        isinstance=isinstance,
        list=list,
        str=str,
        tuple=tuple,
        _intstr=int.__str__,
    ):

    if _indent is not None and not isinstance(_indent, str):
        _indent = ' ' * _indent

    def _iterencode_list(lst, _current_indent_level):
        if not lst:
            yield '[]'
            return
        if markers is not None:
            markerid = id(lst)
            if markerid in markers:
                raise ValueError("Circular reference detected")
            markers[markerid] = lst

        ################ Custom implementation ##################
        def dim(a):
            if isinstance(a, (list, tuple)):
                shape.append(len(a))
                dim(a[0])

        shape = []
        dim(lst)
        # create a new __indent variable and use it for the rest
        # of the _iterencode_list function
        if len(shape) == 1:
            __indent = None
        else:
            __indent = _indent
        #########################################################

        buf = '['
        if __indent is not None:
            _current_indent_level += 1
            newline_indent = '\n' + __indent * _current_indent_level
            separator = _item_separator + newline_indent
            buf += newline_indent
        else:
            newline_indent = None
            separator = _item_separator
        first = True
        for value in lst:
            if first:
                first = False
            else:
                buf = separator
            if isinstance(value, str):
                yield buf + _encoder(value)
            elif value is None:
                yield buf + 'null'
            elif value is True:
                yield buf + 'true'
            elif value is False:
                yield buf + 'false'
            elif isinstance(value, int):
                # Subclasses of int/float may override __str__, but we still
                # want to encode them as integers/floats in JSON. One example
                # within the standard library is IntEnum.
                yield buf + _intstr(value)
            elif isinstance(value, float):
                # see comment above for int
                yield buf + _floatstr(value)
            else:
                yield buf
                if isinstance(value, (list, tuple)):
                    chunks = _iterencode_list(value, _current_indent_level)
                elif isinstance(value, dict):
                    chunks = _iterencode_dict(value, _current_indent_level)
                else:
                    chunks = _iterencode(value, _current_indent_level)
                yield from chunks
        if newline_indent is not None:
            _current_indent_level -= 1
            yield '\n' + __indent * _current_indent_level
        yield ']'
        if markers is not None:
            del markers[markerid]

    def _iterencode_dict(dct, _current_indent_level):
        if not dct:
            yield '{}'
            return
        if markers is not None:
            markerid = id(dct)
            if markerid in markers:
                raise ValueError("Circular reference detected")
            markers[markerid] = dct
        yield '{'
        if _indent is not None:
            _current_indent_level += 1
            newline_indent = '\n' + _indent * _current_indent_level
            item_separator = _item_separator + newline_indent
            yield newline_indent
        else:
            newline_indent = None
            item_separator = _item_separator
        first = True
        if _sort_keys:
            items = sorted(dct.items(), key=lambda kv: kv[0])
        else:
            items = dct.items()
        for key, value in items:
            if isinstance(key, str):
                pass
            # JavaScript is weakly typed for these, so it makes sense to
            # also allow them.  Many encoders seem to do something like this.
            elif isinstance(key, float):
                # see comment for int/float in _make_iterencode
                key = _floatstr(key)
            elif key is True:
                key = 'true'
            elif key is False:
                key = 'false'
            elif key is None:
                key = 'null'
            elif isinstance(key, int):
                # see comment for int/float in _make_iterencode
                key = _intstr(key)
            elif _skipkeys:
                continue
            else:
                raise TypeError('keys must be str, int, float, bool or None, '
                                'not {}'.format(key.__class__.__name__))
            if first:
                first = False
            else:
                yield item_separator
            yield _encoder(key)
            yield _key_separator
            if isinstance(value, str):
                yield _encoder(value)
            elif value is None:
                yield 'null'
            elif value is True:
                yield 'true'
            elif value is False:
                yield 'false'
            elif isinstance(value, int):
                # see comment for int/float in _make_iterencode
                yield _intstr(value)
            elif isinstance(value, float):
                # see comment for int/float in _make_iterencode
                yield _floatstr(value)
            else:
                if isinstance(value, (list, tuple)):
                    chunks = _iterencode_list(value, _current_indent_level)
                elif isinstance(value, dict):
                    chunks = _iterencode_dict(value, _current_indent_level)
                else:
                    chunks = _iterencode(value, _current_indent_level)
                yield from chunks
        if newline_indent is not None:
            _current_indent_level -= 1
            yield '\n' + _indent * _current_indent_level
        yield '}'
        if markers is not None:
            del markers[markerid]

    def _iterencode(o, _current_indent_level):
        if isinstance(o, str):
            yield _encoder(o)
        elif o is None:
            yield 'null'
        elif o is True:
            yield 'true'
        elif o is False:
            yield 'false'
        elif isinstance(o, int):
            # see comment for int/float in _make_iterencode
            yield _intstr(o)
        elif isinstance(o, float):
            # see comment for int/float in _make_iterencode
            yield _floatstr(o)
        elif isinstance(o, (list, tuple)):
            yield from _iterencode_list(o, _current_indent_level)
        elif isinstance(o, dict):
            yield from _iterencode_dict(o, _current_indent_level)
        else:
            if markers is not None:
                markerid = id(o)
                if markerid in markers:
                    raise ValueError("Circular reference detected")
                markers[markerid] = o
            o = _default(o)
            yield from _iterencode(o, _current_indent_level)
            if markers is not None:
                del markers[markerid]
    return _iterencode



import pytest

from msl.io.base import Root
from msl.io.vertex import Vertex


def test_instantiation():
    root = Root('')

    v = Vertex(name='this is ok', parent=root, read_only=True)
    assert v.read_only

    # must specify a name
    with pytest.raises(TypeError):
        Vertex(parent=root, read_only=True)

    # the name must be a non-empty string
    for n in [None, '']:
        with pytest.raises(ValueError, match=r'non-empty string'):
            Vertex(name=n, parent=root, read_only=True)

    # the name cannot contain a '/'
    for n in ['/', '/a', 'a/b']:
        with pytest.raises(ValueError, match=r'cannot contain the "/" character'):
            Vertex(name=n, parent=root, read_only=True)

    # check that the name is forced to be unique
    with pytest.raises(ValueError, match=r'is not unique'):
        Vertex(name='this is ok', parent=root, read_only=True)



import string

import pytest

from msl.io.readers.spreadsheet import Spreadsheet


def test_to_indices():
    to_indices = Spreadsheet.to_indices

    assert to_indices('A0') == (0, 0)
    assert to_indices('A1') == (0, 0)
    assert to_indices('AA2') == (1, 26)
    assert to_indices('AAA123') == (122, 702)
    assert to_indices('XFD123456789') == (123456788, 16383)

    for c in ['', '1', 'A:B', 'A1C10', '1A1']:
        with pytest.raises(ValueError, match='Invalid cell'):
            to_indices(c)
    for c in string.punctuation:
        with pytest.raises(ValueError, match='Invalid cell'):
            to_indices('A'+c)
        with pytest.raises(ValueError, match='Invalid cell'):
            to_indices(c+'A')
        with pytest.raises(ValueError, match='Invalid cell'):
            to_indices('A1'+c)
        with pytest.raises(ValueError, match='Invalid cell'):
            to_indices(c+'A1')

    index = 0
    uppercase = string.ascii_uppercase

    # single letter
    for c in uppercase:
        assert to_indices(c) == (None, index)
        index += 1

    # two letters
    for i in range(26):
        first = uppercase[i]
        for c in uppercase:
            assert to_indices(first+c) == (None, index)
            index += 1

    # three letters
    for i in range(26):
        first = uppercase[i]
        for j in range(26):
            second = uppercase[j]
            for c in uppercase:
                assert to_indices(first + second + c) == (None, index)
                index += 1

    # four letters
    for i in range(26):
        first = uppercase[i]
        for j in range(26):
            second = uppercase[j]
            for k in range(26):
                third = uppercase[k]
                for c in uppercase:
                    assert to_indices(first + second + third + c) == (None, index)
                    index += 1


def test_to_letters():
    to_letters = Spreadsheet.to_letters

    # negative
    for index in [-1, -10, -100, -1000]:
        assert to_letters(index) == ''

    index = 0
    uppercase = string.ascii_uppercase

    # single letter
    for c in uppercase:
        assert to_letters(index) == c
        index += 1

    # two letters
    for i in range(26):
        first = uppercase[i]
        for c in uppercase:
            assert to_letters(index) == first + c
            index += 1

    # three letters
    for i in range(26):
        first = uppercase[i]
        for j in range(26):
            second = uppercase[j]
            for c in uppercase:
                assert to_letters(index) == first + second + c
                index += 1

    # four letters
    for i in range(26):
        first = uppercase[i]
        for j in range(26):
            second = uppercase[j]
            for k in range(26):
                third = uppercase[k]
                for c in uppercase:
                    assert to_letters(index) == first + second + third + c
                    index += 1


def test_to_letters_to_indices():
    to_indices = Spreadsheet.to_indices
    to_letters = Spreadsheet.to_letters
    for index in range(1, int(1e5)):
        assert to_indices(to_letters(index)+str(index)) == (index-1, index)


def test_to_slices():
    to_slices = Spreadsheet.to_slices
    for cells in ['', 'A', 'A:B:', ':']:
        with pytest.raises(ValueError, match='Invalid cell'):
            to_slices(cells)
    assert to_slices('A:A') == (slice(0, None, None), slice(0, 1, None))
    assert to_slices('A:G') == (slice(0, None, None), slice(0, 7, None))
    assert to_slices('A1:A100') == (slice(0, 100, None), slice(0, 1, None))
    assert to_slices('A2:G10', row_step=2, column_step=2) == (slice(1, 10, 2), slice(0, 7, 2))
    assert to_slices('H25:AA47', column_step=2) == (slice(24, 47, None), slice(7, 27, 2))
    assert to_slices('HF11:JK4321', row_step=5) == (slice(10, 4321, 5), slice(213, 271, None))



import logging
import types

import pytest

from msl.io.base import Root
from msl.io.dataset import Dataset
from msl.io.group import Group


def test_instantiation():
    root = Root('some.file')
    assert root.file == 'some.file'
    assert root.name == '/'
    assert not root.read_only
    assert not root.metadata.read_only
    assert len(root) == 0
    assert len(root.metadata) == 0
    assert str(root).startswith('<Root')

    root = Root('C:\\path\\to\\a\\windows.file')
    assert root.file == 'C:\\path\\to\\a\\windows.file'
    assert root.name == '/'

    root = Root(r'\\network\drive with multiple\spa ces.file')
    assert root.file == '\\\\network\\drive with multiple\\spa ces.file'
    assert root.name == '/'

    root = Root('/home/another.xxx')
    assert root.file == '/home/another.xxx'
    assert root.name == '/'
    assert not root.read_only
    assert not root.metadata.read_only
    assert len(root) == 0
    assert len(root.metadata) == 0

    root = Root('/home/another.xxx', one=1, two=2, three=3)
    assert root.file == '/home/another.xxx'
    assert root.name == '/'
    assert not root.read_only
    assert not root.metadata.read_only
    assert len(root) == 0
    assert root.metadata == {'one': 1, 'two': 2, 'three': 3}

    root.read_only = True

    # cannot add metadata
    with pytest.raises(ValueError):
        root.add_metadata(four=4, five=5)

    root.read_only = False
    root.add_metadata(four=4, five=5)
    assert root.metadata == {'one': 1, 'two': 2, 'three': 3, 'four': 4, 'five': 5}


def test_create_group():
    root = Root('')

    # must specify a name for the group
    with pytest.raises(TypeError):
        root.create_group(read_only=True)

    assert not root.read_only
    assert not root.metadata.read_only

    root.read_only = True

    # cannot create a group since root is in read-only mode
    with pytest.raises(ValueError):
        root.create_group('xxx')

    root.read_only = False
    assert not root.read_only
    assert not root.metadata.read_only

    a = root.create_group('a')
    assert root.is_group(a)
    assert a.is_group(a)
    assert isinstance(a, Group)
    assert not a.read_only  # gets read-only value from root
    assert not a.metadata.read_only  # gets read-only value from root
    assert a.name == '/a'
    assert a.parent is root
    assert 'a' in root

    # set read_only=True to create a subgroup that is read only but root is not read only
    b = root.create_group('b', read_only=True)
    assert b.read_only
    assert b.metadata.read_only
    assert not root.read_only
    assert not root.metadata.read_only
    assert b.name == '/b'
    assert b.parent is root
    assert 'b' in root

    # cannot set the parent because this kwarg is pop'd
    c = root.create_group('c', parent=None)
    assert c.name == '/c'
    assert c.parent is root
    assert 'c' in root
    assert len(c.metadata) == 0

    # create a subgroup with some metadata
    d = root.create_group('d', parent=None, one=1, two=2, three=3)
    assert d.name == '/d'
    assert d.parent is root
    assert 'd' in root
    assert 'parent' not in d.metadata
    assert d.metadata == {'one': 1, 'two': 2, 'three': 3}

    # check that we can make root read only again
    root.read_only = True
    with pytest.raises(ValueError):
        root.create_group('xxx')

    # check that the subgroups of root make sense
    assert len(root) == 4
    assert 'a' in root
    assert 'b' in root
    assert 'c' in root
    assert 'd' in root
    assert 'xxx' not in root


def test_create_dataset():
    root = Root('')

    # must specify a name for the dataset
    with pytest.raises(TypeError):
        root.create_dataset()

    assert not root.read_only
    assert not root.metadata.read_only

    root.read_only = True

    # cannot create a dataset if root is in read-only mode
    with pytest.raises(ValueError):
        root.create_dataset('xxx')

    root.read_only = False
    assert not root.read_only
    assert not root.metadata.read_only

    # create an emtpy dataset (no data, no metadata)
    d1 = root.create_dataset('data1')
    assert root.is_dataset(d1)
    assert isinstance(d1, Dataset)
    assert not d1.read_only  # gets read-only value from root
    assert d1.name == '/data1'
    assert d1.parent is root
    assert d1.size == 0
    assert d1.dtype == float
    assert len(d1.metadata) == 0
    assert 'data1' in root

    # create a dataset with shape and metadata
    d2 = root.create_dataset('data2', shape=(10, 5), one=1)
    assert d2.name == '/data2'
    assert d2.parent is root
    assert d2.shape == (10, 5)
    assert d2.size == 50
    assert d2.dtype == float
    assert len(d2.metadata) == 1
    assert d2.metadata == {'one': 1}
    assert 'data2' in root

    # cannot set the parent because this kwarg is pop'd
    d3 = root.create_dataset('data3', parent=None)
    assert d3.name == '/data3'
    assert d3.parent is root
    assert 'data3' in root

    # creating a dataset in read-only mode doesn't change root's read mode
    d4 = root.create_dataset('data4', read_only=True)
    assert d4.read_only
    assert d4.metadata.read_only
    assert not root.read_only
    assert not root.metadata.read_only
    assert 'data4' in root

    # check kwargs
    d5 = root.create_dataset('data5', parent=None, read_only=True, order='F', one=1, two=2, three=3)
    assert d5.name == '/data5'
    assert d5.parent is root
    assert d5.read_only
    assert d5.metadata.read_only
    assert not root.read_only
    assert not root.metadata.read_only
    assert 'data5' in root
    assert 'order' not in d5.metadata
    assert d5.metadata == {'one': 1, 'two': 2, 'three': 3}

    # check that we can make root read only again
    root.read_only = True
    with pytest.raises(ValueError):
        root.create_dataset('xxx')

    # check that the datasets of root make sense
    assert len(root) == 5
    assert 'data1' in root
    assert 'data2' in root
    assert 'data3' in root
    assert 'data4' in root
    assert 'data5' in root
    assert 'xxx' not in root


def test_accessing_subgroups_subdatasets():
    root = Root('')

    a = root.create_group('a')
    d1 = a.create_dataset('d1')
    b = a.create_group('b')
    c = b.create_group('c')
    d = c.create_group('d')
    d2 = d.create_dataset('d2')
    d3 = d.create_dataset('d3')

    assert a is root['/a']
    assert d1 is root['/a/d1']
    assert b is root['/a/b']
    assert c is root['/a/b/c']
    assert d is root['/a/b/c/d']
    assert d2 is root['/a/b/c/d/d2']
    assert d3 is root['/a/b/c/d/d3']

    assert d1 is a['/d1']
    assert b is a['/b']
    assert c is a['/b/c']
    assert d is a['/b/c/d']
    assert d2 is a['/b/c/d/d2']
    assert d3 is a['/b/c/d/d3']

    assert c is b['/c']
    assert d is b['/c/d']
    assert d2 is b['/c/d/d2']
    assert d3 is b['/c/d/d3']

    assert d is c['/d']
    assert d2 is c['/d/d2']
    assert d3 is c['/d/d3']

    assert d2 is d['/d2']
    assert d3 is d['/d3']

    assert root.a is root['a']
    assert root.a.d1 is root['a']['d1']
    assert root.a.b is root['a']['b']
    assert root.a.b.c is root['a']['b']['c']
    assert root.a.b.c.d is root['a']['b']['c']['d']
    assert root.a.b.c.d.d2 is root['a']['b']['c']['d']['d2']
    assert root.a.b.c.d.d3 is root['a']['b']['c']['d']['d3']

    assert root['a'].b['c'].d is root['/a/b/c/d']

    assert root.a.d1 is a['d1']
    assert root.a.b is a['b']
    assert root.a.b.c is b['c']
    assert root.a.b.c.d is c['d']
    assert root.a.b.c.d.d2 is d['d2']

    aa = root.a
    assert aa.b is root.a.b
    assert aa.b is root['a']['b']

    cc = root.a.b.c
    assert cc.d is root.a.b.c.d
    assert cc['d'] is root['a']['b']['c']['d']

    with pytest.raises(KeyError):
        xxx = root['xxx']

    with pytest.raises(AttributeError):
        xxx = root.xxx


def test_in_not_in():
    root = Root('')

    a = root.create_group('a')
    a.create_dataset('first dataset')
    b = a.create_group('b')
    c = b.create_group('c')
    d = c.create_group('d')
    d.create_dataset('second dataset')

    assert 'xxx' not in root
    assert '/' not in root
    assert 'a' in root
    assert '/a' in root
    assert 'first dataset' in a
    assert '/a/first dataset' in root
    assert 'first dataset' not in root
    assert 'b' in a
    assert '/b' in a
    assert '/a/b' in root
    assert 'c' in b
    assert '/c' in b
    assert '/a/b/c' in root
    assert '/c' not in a
    assert 'd' in c
    assert '/d' in c
    assert 'second dataset' in d
    assert '/a/b/c/d/second dataset' in root


def test_read_only_propagates():
    root = Root('')

    g1 = root.create_group('g1')
    d1 = g1.create_dataset('d1')
    g2 = g1.create_group('g2')
    g3 = g2.create_group('g3')
    d3 = g3.create_dataset('d3')
    g4 = g3.create_group('g4')

    # all sub groups/datasets inherit roots read-only value
    assert not root.read_only
    assert not root.metadata.read_only
    assert not g1.read_only
    assert not g1.metadata.read_only
    assert not d1.read_only
    assert not d1.metadata.read_only
    assert not g2.read_only
    assert not g2.metadata.read_only
    assert not g3.read_only
    assert not g3.metadata.read_only
    assert not d3.read_only
    assert not d3.metadata.read_only
    assert not g4.read_only
    assert not g4.metadata.read_only

    # make all sub groups/datasets read only by only changing root
    root.read_only = True
    assert root.read_only
    assert root.metadata.read_only
    assert g1.read_only
    assert g1.metadata.read_only
    assert d1.read_only
    assert d1.metadata.read_only
    assert g2.read_only
    assert g2.metadata.read_only
    assert g3.read_only
    assert g3.metadata.read_only
    assert d3.read_only
    assert d3.metadata.read_only
    assert g4.read_only
    assert g4.metadata.read_only

    # make all sub groups/datasets <= g2 writeable
    g2.read_only = False
    assert root.read_only
    assert root.metadata.read_only
    assert g1.read_only
    assert g1.metadata.read_only
    assert d1.read_only
    assert d1.metadata.read_only
    assert not g2.read_only
    assert not g2.metadata.read_only
    assert not g3.read_only
    assert not g3.metadata.read_only
    assert not d3.read_only
    assert not d3.metadata.read_only
    assert not g4.read_only
    assert not g4.metadata.read_only


def test_datasets_groups():
    root = Root('')

    d0 = root.create_dataset('d0')
    g1 = root.create_group('g1')
    d1 = g1.create_dataset('d1')
    g2 = g1.create_group('g2')
    g3 = g2.create_group('g3')
    d3 = g3.create_dataset('d3')
    g4 = g3.create_group('g4')

    # cannot create 2 sub-Groups with the same key
    with pytest.raises(ValueError):
        root.create_group('g1')

    # cannot create 2 Datasets with the same key
    with pytest.raises(ValueError):
        g3.create_group('d3')

    assert isinstance(root.datasets(), types.GeneratorType)
    datasets = list(root.datasets())
    assert len(datasets) == 3
    assert d0 in datasets
    assert d1 in datasets
    assert d3 in datasets

    assert isinstance(root.groups(), types.GeneratorType)
    groups = list(root.groups())
    assert len(groups) == 4
    assert g1 in groups
    assert g2 in groups
    assert g3 in groups
    assert g4 in groups

    root_items = {
        '/d0': d0,
        '/g1': g1,
        '/g1/d1': d1,
        '/g1/g2': g2,
        '/g1/g2/g3': g3,
        '/g1/g2/g3/d3': d3,
        '/g1/g2/g3/g4': g4,
    }
    for key, value in root.items():
        assert key in root_items
        assert value is root_items[key]
        del root_items[key]


def test_delete_vertex():
    root = Root('')

    root.create_group('g1')
    g2 = root.create_group('g2')
    g2.create_group('a')
    g2.create_dataset('b')
    c = g2.create_group('c')
    root.create_group('g3')
    c.create_dataset('cd1')
    c.create_dataset('cd2')
    c.create_group('cg3')

    with pytest.raises(KeyError):  # invalid key
        del root['x']

    with pytest.raises(KeyError):  # invalid key
        del root['x']['y']

    with pytest.raises(AttributeError):  # invalid attribute
        del root.x

    with pytest.raises(AttributeError):  # invalid attribute
        del root.x.y

    root.read_only = True

    with pytest.raises(ValueError):  # read-only mode
        del root['g1']

    with pytest.raises(ValueError):  # read-only mode
        del root.g2

    with pytest.raises(ValueError):  # read-only mode
        del root.g2.c.cd1

    assert '/g1' in root
    assert '/g2' in root
    assert '/g2/a' in root
    assert '/g2/b' in root
    assert '/g2/c' in root
    assert '/g2/c/cd1' in root
    assert '/g2/c/cd2' in root
    assert '/g2/c/cg3' in root
    assert '/g3' in root

    root.read_only = False

    del root['g1']
    assert '/g1' not in root
    assert '/g2' in root
    assert '/g2/a' in root
    assert '/g2/b' in root
    assert '/g2/c' in root
    assert '/g2/c/cd1' in root
    assert '/g2/c/cd2' in root
    assert '/g2/c/cg3' in root
    assert '/g3' in root

    root.read_only = True

    with pytest.raises(ValueError):  # read-only mode
        del root['g2']

    with pytest.raises(ValueError):  # read-only mode
        del root.g2

    with pytest.raises(ValueError):  # read-only mode
        del root.g2.c.cg3

    root.read_only = False

    del root['/g2/a']
    assert '/g2' in root
    assert '/g2/a' not in root
    assert '/g2/b' in root
    assert '/g2/c' in root
    assert '/g2/c/cd1' in root
    assert '/g2/c/cd2' in root
    assert '/g2/c/cg3' in root
    assert '/g3' in root

    del root['g2']['c']['cg3']
    assert '/g2' in root
    assert '/g2/b' in root
    assert '/g2/c' in root
    assert '/g2/c/cd1' in root
    assert '/g2/c/cd2' in root
    assert '/g2/c/cg3' not in root
    assert '/g3' in root

    del root.g2.b
    assert '/g2' in root
    assert '/g2/b' not in root
    assert '/g2/c' in root
    assert '/g2/c/cd1' in root
    assert '/g2/c/cd2' in root
    assert '/g3' in root

    del root['/g2']
    assert '/g2' not in root
    assert '/g2/c' not in root
    assert '/g2/c/cd1' not in root
    assert '/g2/c/cd2' not in root
    assert '/g3' in root

    del root['g3']
    assert len(root) == 0


def test_auto_create_subgroups():
    root = Root('')

    assert len(list(root.groups())) == 0
    assert len(list(root.datasets())) == 0

    root.create_group('a/group2/c/group4/d/group6')
    root.create_dataset('/w/x/y/z', shape=(10,))

    # intermediate Groups get created automatically
    with pytest.raises(ValueError):
        root.create_group('a/group2/c')
    with pytest.raises(ValueError):
        root.create_group('/w/x')

    assert len(list(root.groups())) == 9

    root.a.group2.c.group4.create_group('/m/n')

    assert len(list(root.groups())) == 11

    assert 'a' in root
    assert 'group2' in root.a
    assert 'c' in root.a.group2
    assert 'group4' in root.a.group2.c
    assert 'm' in root.a.group2.c.group4
    assert 'm/n' in root.a.group2.c.group4
    assert 'n' in root.a.group2.c.group4.m
    assert 'd' in root.a.group2.c.group4
    assert 'group6' in root.a.group2.c.group4.d

    assert 'w' in root
    assert 'w/x' in root
    assert 'w/x/y' in root
    assert 'w/x/y/z' in root
    assert root.is_dataset(root['/w/x/y/z'])
    assert root['/w/x/y/z'].shape == (10,)


def test_requires():
    root = Root('')

    #
    # Groups
    #
    assert 'a' not in root

    a = root.require_group('a')
    assert root.is_group(a)
    assert 'a' in root
    assert '/a' in root
    assert root.require_group('a') is a
    assert root.require_group('/a') is a

    # group exists but adding new metadata to it
    a2 = root.require_group('a', one=1)
    assert a2 is a
    assert a.metadata.one == 1
    assert a2.metadata.one == 1

    # try to add Metadata to a Group that is read only
    root.read_only = True
    with pytest.raises(ValueError):
        root.require_group('a', two=2)
    # read-only mode
    with pytest.raises(ValueError):
        a.create_group('b')

    root.read_only = False
    b = a.create_group('b')
    with pytest.raises(ValueError):
        a.create_group('b')
    assert root.require_group('a/b') is b

    root.require_group('/a/b/c/d/e/', foo='bar')
    assert 'a' in root
    assert 'b' in a
    assert 'c' in root.a.b
    assert 'd' in root.a.b.c
    assert 'foo' not in root.metadata
    assert 'foo' not in root.a.metadata
    assert 'foo' not in root.a.b.metadata
    assert 'foo' not in root.a.b.c.metadata
    assert 'foo' not in root.a.b.c.d.metadata
    assert 'foo' in root.a.b.c.d.e.metadata
    assert root.a.b.c.d.e.metadata.foo == 'bar'

    # make sure that calling require_group from a sub-Group works properly
    c = a.b.c
    d = c.d
    e = d.e
    assert a.require_group('b') is b
    assert a.require_group('/b/c') is c
    assert a.require_group('b/c/') is root.a.b.c
    assert b.require_group('c/d/e') is e
    assert b.require_group('c/d/e') is root.a.b.c.d.e
    assert c.require_group('d') is d
    assert c.require_group('d') is root.a.b.c.d
    assert c.require_group('d/e') is e
    assert c.require_group('/d/e/') is root.a.b.c.d.e
    assert d.require_group('e') is e
    assert d.require_group('e/') is root.a.b.c.d.e

    # change the read-only value of the new sub-groups that are required
    assert not root.read_only
    bb = root.require_group('aa/bb', read_only=True, hello='world')
    assert bb is root.aa.bb
    assert not root.read_only
    assert not root.metadata.read_only
    assert root.aa.read_only
    assert root.aa.metadata.read_only
    assert bb.read_only
    assert bb.metadata.read_only
    assert len(root.aa.bb.metadata) == 1
    assert root.aa.bb.metadata.hello == 'world'
    with pytest.raises(ValueError):
        bb.add_metadata(one=1)
    bb.read_only = False
    bb.add_metadata(one=1)
    assert len(root.aa.bb.metadata) == 2
    assert root.aa.bb.metadata.hello == 'world'
    assert root.aa.bb.metadata.one == 1
    assert root.aa.read_only
    assert root.aa.metadata.read_only
    with pytest.raises(ValueError):
        root.aa.add_metadata(two=2)

    # require root.aa.bb but change the read-only value
    assert not root.aa.bb.read_only
    bb2 = root.require_group('aa/bb', read_only=True)
    assert bb2 is root.aa.bb
    assert bb2.read_only
    assert root.aa.bb.read_only
    with pytest.raises(ValueError):
        bb2.add_metadata(three=3)

    #
    # Datasets
    #
    with pytest.raises(ValueError):
        root.require_dataset('a')  # 'a' is already a Group but we are creating a Dataset
    w = root.require_dataset('w')
    assert root.is_dataset(w)
    assert 'w' in root
    assert '/w' in root
    assert root.require_dataset('w') is w
    assert root.require_dataset('/w/') is w

    # dataset exists but adding new metadata to it
    w2 = root.require_dataset('w', one=1)
    assert w2 is w
    assert len(w2.metadata) == 1
    assert w.metadata.one == 1
    assert w2.metadata.one == 1

    # dataset exists but ignores key-value pairs that are not Metadata but are used to create the dataset
    w2 = root.require_dataset('w', shape=(10,), order=None)
    assert w2 is w
    assert len(w2.metadata) == 1
    assert w.metadata.one == 1
    assert w2.metadata.one == 1
    assert 'shape' not in w2.metadata
    assert 'order' not in w2.metadata

    # try to add Metadata to a Dataset that is read only
    root.read_only = True
    with pytest.raises(ValueError):
        root.require_dataset('w', two=2)
    # read-only mode
    with pytest.raises(ValueError):
        root.create_dataset('x')

    # add a Dataset to the 'a' Group
    root.read_only = False
    x = a.create_dataset('x')
    with pytest.raises(ValueError):
        a.create_dataset('x')
    assert root.require_dataset('/a/x') is x

    # add a Dataset to the 'b' Group, create the necessary sub-Groups automatically
    root.require_dataset('/b/x/y/z', data=[1, 2, 3, 4], foo='bar')
    assert root.is_group(root.b)
    assert root.is_group(root.b.x)
    assert root.is_group(root.b.x.y)
    assert root.is_dataset(root.b.x.y.z)
    assert 'w' in root
    assert 'b' in root
    assert 'x' in root.b
    assert 'y' in root.b.x
    assert 'z' in root.b.x.y
    assert 'foo' not in root.metadata
    assert 'foo' not in root.b.metadata
    assert 'foo' not in root.b.x.metadata
    assert 'foo' not in root.b.x.y.metadata
    assert 'foo' in root.b.x.y.z.metadata
    assert root.b.x.y.z.metadata.foo == 'bar'
    assert root.b.x.y.z.shape == (4,)
    assert root.b.x.y.z.tolist() == [1, 2, 3, 4]
    assert root.b.x.y.z.max() == 4

    # change the read-only value of the new Dataset that is required
    assert not root.read_only
    yy = root.require_dataset('xx/yy', read_only=True, hello='world')
    assert yy is root.xx.yy
    assert not root.read_only
    assert not root.metadata.read_only
    assert root.xx.read_only
    assert root.xx.metadata.read_only
    assert yy.read_only
    assert yy.metadata.read_only
    assert len(root.xx.yy.metadata) == 1
    assert root.xx.yy.metadata.hello == 'world'
    with pytest.raises(ValueError):
        yy.add_metadata(one=1)
    yy.read_only = False
    yy.add_metadata(one=1)
    assert len(root.xx.yy.metadata) == 2
    assert root.xx.yy.metadata.hello == 'world'
    assert root.xx.yy.metadata.one == 1
    assert root.xx.read_only
    assert root.xx.metadata.read_only
    with pytest.raises(ValueError):
        root.xx.add_metadata(two=2)

    # require root.xx.yy but change the read-only value
    assert not root.xx.yy.read_only
    yy2 = root.require_dataset('/xx/yy/', read_only=True)
    assert yy2 is root.xx.yy
    assert yy2.read_only
    assert root.xx.yy.read_only
    with pytest.raises(ValueError):
        yy2.add_metadata(three=3)

    # make sure that calling require_dataset from a sub-Group works properly
    a = root.a
    assert a.require_dataset('x') is x
    assert a.require_dataset('/x') is root.a.x
    b = root['b']
    assert b.require_dataset('x/y/z') is root.b.x.y.z
    x = root['b']['x']
    assert x.require_dataset('/y/z') is root.b.x.y.z
    y = root['b']['x']['y']
    assert y.require_dataset('/z/') is root.b.x.y.z
    assert root['b']['x']['y'].require_dataset('z') is root.b.x.y.z
    xx = root.xx
    assert xx.require_dataset('yy') is root['xx']['yy']
    assert root.xx.require_dataset('yy/') is yy2


def test_tree():
    root = Root('')
    a = root.create_group('a')
    a.create_dataset('d1')
    b = a.create_group('b')
    b.create_dataset('d2')
    c = b.create_group('c')
    root.create_group('x/y/z')
    d = c.create_group('d')
    a.create_dataset('d3')
    root.create_dataset('d4')
    d.create_dataset('d5')
    d.create_dataset('d6')
    root.create_dataset('d7')

    tree = """
<Root '' (7 groups, 7 datasets, 0 metadata)>
  <Group '/a' (3 groups, 5 datasets, 0 metadata)>
    <Group '/a/b' (2 groups, 3 datasets, 0 metadata)>
      <Group '/a/b/c' (1 groups, 2 datasets, 0 metadata)>
        <Group '/a/b/c/d' (0 groups, 2 datasets, 0 metadata)>
          <Dataset '/a/b/c/d/d5' shape=(0,) dtype='<f8' (0 metadata)>
          <Dataset '/a/b/c/d/d6' shape=(0,) dtype='<f8' (0 metadata)>
      <Dataset '/a/b/d2' shape=(0,) dtype='<f8' (0 metadata)>
    <Dataset '/a/d1' shape=(0,) dtype='<f8' (0 metadata)>
    <Dataset '/a/d3' shape=(0,) dtype='<f8' (0 metadata)>
  <Dataset '/d4' shape=(0,) dtype='<f8' (0 metadata)>
  <Dataset '/d7' shape=(0,) dtype='<f8' (0 metadata)>
  <Group '/x' (2 groups, 0 datasets, 0 metadata)>
    <Group '/x/y' (1 groups, 0 datasets, 0 metadata)>
      <Group '/x/y/z' (0 groups, 0 datasets, 0 metadata)>"""

    # Python 2.7 64-bit has shape=(0L,) and we don't care about (0L,) vs (0,)
    assert root.tree().replace('shape=(0L,)', 'shape=(0,)') == tree[1:]  # skip the first line

    # use del instead of Group.remove()
    del root.a.b.c

    tree = """
<Root '' (5 groups, 5 datasets, 0 metadata)>
  <Group '/a' (1 groups, 3 datasets, 0 metadata)>
    <Group '/a/b' (0 groups, 1 datasets, 0 metadata)>
      <Dataset '/a/b/d2' shape=(0,) dtype='<f8' (0 metadata)>
    <Dataset '/a/d1' shape=(0,) dtype='<f8' (0 metadata)>
    <Dataset '/a/d3' shape=(0,) dtype='<f8' (0 metadata)>
  <Dataset '/d4' shape=(0,) dtype='<f8' (0 metadata)>
  <Dataset '/d7' shape=(0,) dtype='<f8' (0 metadata)>
  <Group '/x' (2 groups, 0 datasets, 0 metadata)>
    <Group '/x/y' (1 groups, 0 datasets, 0 metadata)>
      <Group '/x/y/z' (0 groups, 0 datasets, 0 metadata)>"""

    # Python 2.7 64-bit has shape=(0L,) and we don't care about (0L,) vs (0,)
    assert root.tree().replace('shape=(0L,)', 'shape=(0,)') == tree[1:]  # skip the first line

    # use Group.remove() instead of del
    root.remove('a')

    tree = """
<Root '' (3 groups, 2 datasets, 0 metadata)>
  <Dataset '/d4' shape=(0,) dtype='<f8' (0 metadata)>
  <Dataset '/d7' shape=(0,) dtype='<f8' (0 metadata)>
  <Group '/x' (2 groups, 0 datasets, 0 metadata)>
    <Group '/x/y' (1 groups, 0 datasets, 0 metadata)>
      <Group '/x/y/z' (0 groups, 0 datasets, 0 metadata)>"""

    # Python 2.7 64-bit has shape=(0L,) and we don't care about (0L,) vs (0,)
    assert root.tree().replace('shape=(0L,)', 'shape=(0,)') == tree[1:]  # skip the first line

    # increase the indentation
    tree = """
<Root '' (3 groups, 2 datasets, 0 metadata)>
     <Dataset '/d4' shape=(0,) dtype='<f8' (0 metadata)>
     <Dataset '/d7' shape=(0,) dtype='<f8' (0 metadata)>
     <Group '/x' (2 groups, 0 datasets, 0 metadata)>
          <Group '/x/y' (1 groups, 0 datasets, 0 metadata)>
               <Group '/x/y/z' (0 groups, 0 datasets, 0 metadata)>"""

    # Python 2.7 64-bit has shape=(0L,) and we don't care about (0L,) vs (0,)
    assert root.tree(indent=5).replace('shape=(0L,)', 'shape=(0,)') == tree[1:]  # skip the first line


def test_add_group():
    root = Root('some file')

    for item in [dict(), tuple(), list(), None, Dataset('dset', None, False)]:
        with pytest.raises(TypeError):
            root.add_group('name', item)

    root2 = Root('New')
    assert len(root2) == 0

    #
    # add a Group that does not contain sub-Groups nor Datasets
    #

    # add to the Root Group
    root2.add_group('', root.create_group('a', one=1, foo='bar'))
    assert len(root2) == 1
    assert root2.a is not root.a
    assert '/a' in root2
    assert len(root2.a.metadata) == 2
    assert root2.a.metadata.one == 1
    assert root2.a.metadata['foo'] == 'bar'

    root2.clear()  # also tests Root.clear()
    assert len(root2) == 0

    # creates an "/B" Group and then add to it
    root2.add_group('B', root.create_group('b', two=2))
    assert len(root2) == 2
    assert root2.B.b is not root.b
    assert '/B/b' in root2
    assert 'B' in root2
    assert 'b' in root2.B
    assert '/B/b' in root2
    assert len(root2.B.metadata) == 0
    assert len(root2.B.b.metadata) == 1
    assert root2.B.b.metadata.two == 2

    root2.clear()
    assert len(root2) == 0

    # creates an "/A/B/C" Group and then add to it (add a ridiculous amount of '/')
    root2.add_group('/////A/B/C//////////', root.create_group('c', x='x', y='y'))
    assert len(root2) == 4
    assert root2.A.B.C.c is not root.c
    assert '/A' in root2
    assert 'A/B' in root2
    assert '/A/B/C' in root2
    assert '/A/B/C/c' in root2
    assert '/c' in root2.A.B.C
    assert len(root2.A.metadata) == 0
    assert len(root2.A.B.metadata) == 0
    assert len(root2.A.B.C.metadata) == 0
    assert len(root2.A.B.C.c.metadata) == 2
    assert root2.A.B.C.c.metadata.x == 'x'
    assert root2['A']['B'].C['c'].metadata['y'] == 'y'

    # verify root's tree
    assert len(root) == 3
    assert 'a' in root
    assert '/b' in root
    assert 'c' in root
    assert len(root.a.metadata) == 2
    assert root.a.metadata.one == 1
    assert root.a.metadata.foo == 'bar'
    assert len(root.b.metadata) == 1
    assert root.b.metadata.two == 2
    assert len(root.c.metadata) == 2
    assert root.c.metadata.x == 'x'
    assert root.c.metadata.y == 'y'

    # add some Datasets to root
    root.b.create_dataset('/x', data=[0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    root.c.create_dataset('y/z', shape=(3, 4), meta='data')
    assert 'x' in root.b
    assert 'z' in root.c.y

    # add root to root2
    root2.add_group('/old', root)
    assert len(root2) == 11
    assert '/A' in root2
    assert 'A/B' in root2
    assert '/A/B/C' in root2
    assert '/A/B/C/c' in root2
    assert '/c' in root2.A.B.C
    assert 'old' in root2
    assert 'old/a' in root2
    assert '/old/b' in root2
    assert '/old/c' in root2
    assert '/old/b/x' in root2
    assert 'y' in root2.old.c
    assert '/y/z' in root2.old.c
    assert len(root2.A.metadata) == 0
    assert len(root2.A.B.metadata) == 0
    assert len(root2.A.B.C.metadata) == 0
    assert len(root2.A.B.C.c.metadata) == 2
    assert root2.A.B.C.c.metadata.x == 'x'
    assert root2['A']['B'].C['c'].metadata['y'] == 'y'
    assert len(root2.old.a.metadata) == 2
    assert root2.old.a.metadata.one == 1
    assert root2.old.a.metadata.foo == 'bar'
    assert len(root2.old.b.metadata) == 1
    assert root2.old.b.metadata.two == 2
    assert len(root2.old.c.metadata) == 2
    assert root2.old.c.metadata.x == 'x'
    assert root2.old.c.metadata.y == 'y'
    assert len(root2.old.b.metadata) == 1
    assert root2.old.b.metadata.two == 2
    assert len(root2.old.c.metadata) == 2
    assert root2.old.c.metadata.x == 'x'
    assert root2.old['c'].metadata['y'] == 'y'
    assert len(root2.old.c.y.metadata) == 0
    assert len(root2.old.c.y.z.metadata) == 1
    assert root2.old.c.y.z.metadata.meta == 'data'
    assert root2.old.b.x.shape == (10,)
    assert root2.old.c.y.z.shape == (3, 4)

    # the Metadata is a copy
    root2.old.c.y.z.metadata.meta = 'new value'
    assert root2.old.c.y.z.metadata.meta is not root.c.y.z.metadata.meta
    assert root2.old.c.y.z.metadata.meta == 'new value'
    assert root.c.y.z.metadata.meta == 'data'

    # the data in the Dataset is a copy
    assert root2.old.b.x.data is not root.b.x.data
    assert root2.old.c.y.z.data is not root.c.y.z.data
    root2.old.b.x[:] = 1
    assert sum(root2.old.b.x.data) == 10
    for val in root.b.x.data.tolist():
        assert val == 0


def test_add_dataset():
    root = Root('some file')
    for item in [dict(), tuple(), list(), None, 1.0, Root('r'), Group('group', None, True)]:
        with pytest.raises(TypeError):
            root.add_dataset('name', item)

    assert len(root) == 0

    dset1 = Dataset('dset1', None, False, data=[1, 2, 3], dtype=int)
    dset2 = Dataset('dset2', None, False, data=[4.0, 5.0, 6.0], foo='bar')
    dset3 = Dataset('dset3', None, False, data=[-23.4, 1.78], one=1, two=2)

    root.add_dataset('dset1_copy', dset1)

    d = root.create_group('a/b/c/d')
    c = d.parent
    assert c is root.a.b.c
    c.add_dataset('dset2_copy', dset2)

    c.add_dataset('/x/y/dset3_copy', dset3)

    assert len(list(root.datasets())) == 3
    assert len(list(root.a.datasets())) == 2
    assert len(list(root.a.b.datasets())) == 2
    assert len(list(root.a.b.c.datasets())) == 2
    assert len(list(root.a.b.c.x.datasets())) == 1
    assert len(list(root.a.b.c.x.y.datasets())) == 1
    assert len(list(root.a.b.c.d.datasets())) == 0

    assert root.dset1_copy is not dset1
    assert root.a.b.c.dset2_copy is not dset2
    assert root.a.b.c.x.y.dset3_copy is not dset3

    assert all(v1 == v2 for v1, v2 in zip(dset1, root.dset1_copy))
    assert all(v1 == v2 for v1, v2 in zip(dset2, root.a.b.c.dset2_copy))
    assert all(v1 == v2 for v1, v2 in zip(dset3, root.a.b.c.x.y.dset3_copy))

    assert len(root.dset1_copy.metadata) == 0
    assert len(c.dset2_copy.metadata) == 1
    assert c.dset2_copy.metadata.foo == 'bar'
    assert len(c.x.y.dset3_copy.metadata) == 2
    assert c.x.y.dset3_copy.metadata['one'] == 1
    assert c['x']['y']['dset3_copy'].metadata.two == 2


def test_add_dataset_logging():
    num_initial_handlers = len(logging.getLogger().handlers)

    logger = logging.getLogger('test_add_dataset_logging')

    root = Root('some file')
    for item in [dict(), tuple(), list(), None, 1.0, Root('r'),
                 Group('group', None, True), Dataset('d', None, True)]:
        with pytest.raises(TypeError):
            root.add_dataset_logging('name', item)

    assert len(root) == 0
    assert len(logger.handlers) == 0
    assert len(logging.getLogger().handlers) == num_initial_handlers

    dsetlog1 = root.create_dataset_logging('dsetlog1', level='DEBUG', attributes=['levelname', 'message'],
                                           logger=logger, date_fmt='%H-%M')

    logger.info('bang!')

    root2 = Root('some file')
    root2.create_group('a/b/c')
    b = root2.a.b
    b.add_dataset_logging('x/y/dsetlog2', dsetlog1)

    assert len(logger.handlers) == 2
    assert len(logging.getLogger().handlers) == num_initial_handlers

    logger.debug('hello')
    logger.info('world')

    dsetlog1.remove_handler()
    assert len(logger.handlers) == 1
    assert len(logging.getLogger().handlers) == num_initial_handlers

    logger.warning('foo')
    logger.error('bar')

    assert dsetlog1.metadata['logging_level'] == logging.DEBUG
    assert dsetlog1.metadata.logging_level_name == 'DEBUG'
    assert dsetlog1.metadata.logging_date_format == '%H-%M'
    assert dsetlog1.level == logging.DEBUG
    assert dsetlog1.date_fmt == '%H-%M'
    assert dsetlog1.logger is logger
    assert all(a == b for a, b in zip(dsetlog1.attributes, ['levelname', 'message']))

    assert b.x.y.dsetlog2.metadata['logging_level'] == logging.DEBUG
    assert b.x.y.dsetlog2.metadata.logging_level_name == 'DEBUG'
    assert b.x.y.dsetlog2.metadata.logging_date_format == '%H-%M'
    assert b.x.y.dsetlog2.level == logging.DEBUG
    assert b.x.y.dsetlog2.date_fmt == '%H-%M'
    assert b.x.y.dsetlog2.logger is logger
    assert all(a == b for a, b in zip(dsetlog1.attributes, b.x.y.dsetlog2.attributes))

    assert root.dsetlog1.data.size == 3
    row = root.dsetlog1.data[0]
    assert row['levelname'] == 'INFO' and row['message'] == 'bang!'
    row = root.dsetlog1.data[1]
    assert row['levelname'] == 'DEBUG' and row['message'] == 'hello'
    row = root.dsetlog1.data[2]
    assert row['levelname'] == 'INFO' and row['message'] == 'world'

    assert root2.a.b.x.y.dsetlog2.data.size == 5
    y = b.x.y
    row = root.dsetlog1.data[0]
    assert row['levelname'] == 'INFO' and row['message'] == 'bang!'
    row = y.dsetlog2.data[1]
    assert row['levelname'] == 'DEBUG' and row['message'] == 'hello'
    row = y.dsetlog2.data[2]
    assert row['levelname'] == 'INFO' and row['message'] == 'world'
    row = y.dsetlog2.data[3]
    assert row['levelname'] == 'WARNING' and row['message'] == 'foo'
    row = y.dsetlog2.data[4]
    assert row['levelname'] == 'ERROR' and row['message'] == 'bar'

    root2.a.b.x.y.dsetlog2.remove_handler()

    assert len(logger.handlers) == 0
    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_remove():
    root = Root('')
    a = root.create_group('a')
    a.create_dataset('d1')
    b = a.create_group('b')
    b.create_dataset('d2')
    c = b.create_group('c')
    z = root.create_group('x/y/z')
    d = c.create_group('d')
    a.create_dataset('d3')
    root.create_dataset('d4')
    d.create_dataset('d5')
    d.create_dataset('d6')
    d7 = root.create_dataset('d7')

    assert len(root) == 14
    assert len(list(root.groups())) == 7
    assert len(list(root.datasets())) == 7
    assert 'a' in root
    assert 'd1' in root.a
    assert 'b' in root.a
    assert 'd2' in root.a.b
    assert 'c' in root.a.b
    assert 'x' in root
    assert 'x/y' in root
    assert 'y' in root.x
    assert 'x/y/z' in root
    assert 'z' in root.x.y
    assert 'd' in root.a.b.c
    assert '/a/d3' in root
    assert 'd4' in root
    assert '/d5' in root.a.b.c.d
    assert '/a/b/c/d/d6' in root
    assert 'd7' in root

    # remove the 'd7' Dataset
    d7_2 = root.remove('d7')
    assert len(root) == 13
    assert len(list(root.groups())) == 7
    assert len(list(root.datasets())) == 6
    assert 'd7' not in root
    assert d7_2 is d7

    # remove the 'z' Group
    assert root.remove('z') is None
    assert len(root) == 13
    assert 'x/y/z' in root
    assert root.remove('/y/z') is None
    assert len(root) == 13
    assert 'x/y/z' in root
    z2 = root.x.remove('y/z')
    assert z2 is z
    assert len(root) == 12
    assert len(list(root.groups())) == 6
    assert len(list(root.datasets())) == 6
    assert '/x/y/z' not in root

    # cannot remove in read-only mode
    root.read_only = True
    with pytest.raises(ValueError):
        root.remove('a')
    assert len(root) == 12
    assert 'a' in root

    # remove Group 'd' (which also removes the 'd5' and 'd6' Datasets)
    root.a.b.c.read_only = False
    d2 = root.a.b.c.remove('d')
    assert len(root) == 9
    assert len(root.a) == 5
    assert len(list(root.a.groups())) == 2
    assert len(list(root.a.datasets())) == 3
    assert len(root.a.b) == 2
    assert len(list(root.a.b.groups())) == 1
    assert len(list(root.a.b.datasets())) == 1
    assert 'd' not in root.a.b.c
    assert '/d' not in root.a.b.c
    assert '/a/b/c/d' not in root
    assert '/b/c/d' not in root.a
    assert '/c/d' not in root.a.b
    assert 'd/d5' not in root.a.b.c
    assert 'd/d6' not in root.a.b.c
    assert 'c/d/d5' not in root.a.b
    assert 'c/d/d6' not in root.a.b
    assert 'b/c/d/d5' not in root.a
    assert 'b/c/d/d6' not in root.a
    assert 'a/b/c/d/d5' not in root
    assert 'a/b/c/d/d6' not in root
    assert d2 is d
    with pytest.raises(ValueError):  # root.a.b is still in read-only mode
        root.a.b.remove('c')

    # remove Group 'a'
    root.read_only = False
    a2 = root.remove('a')
    assert len(root) == 3
    assert len(list(root.groups())) == 2
    assert len(list(root.datasets())) == 1
    assert a2 is a
    assert 'a' not in root
    assert 'd4' in root
    assert 'x' in root
    assert '/x/y' in root
    assert 'y' in root.x

    root.clear()
    assert len(root) == 0

    # Check that removing an Group/Dataset also removes it from the descendants
    root = Root('')
    root.require_group('g1/g2/g3/g4')
    root.require_dataset('g1/d0')
    d1 = root.require_dataset('g1/g2/d1')
    root.require_dataset('g1/g2/d2')
    d3 = root.require_dataset('g1/g2/g3/d3')
    root.require_dataset('g1/g2/g3/g4/d4')

    assert len(root) == 9
    assert len(list(root.groups())) == 4
    assert len(list(root.datasets())) == 5
    assert len(root.g1) == 8
    assert len(list(root.g1.groups())) == 3
    assert len(list(root.g1.datasets())) == 5
    assert len(root.g1.g2) == 6
    assert len(list(root.g1.g2.groups())) == 2
    assert len(list(root.g1.g2.datasets())) == 4
    assert len(root.g1.g2.g3) == 3
    assert len(list(root.g1.g2.g3.groups())) == 1
    assert len(list(root.g1.g2.g3.datasets())) == 2
    assert len(root.g1.g2.g3.g4) == 1
    assert len(list(root.g1.g2.g3.g4.groups())) == 0
    assert len(list(root.g1.g2.g3.g4.datasets())) == 1

    d3_removed = root.remove('g1/g2/g3/d3')
    assert d3_removed is d3
    assert 'd3' not in root['/g1/g2/g3']
    assert 'd3' not in root.g1['g2/g3']
    assert 'd3' not in root.g1.g2['/g3']
    assert len(root) == 8
    assert len(list(root.groups())) == 4
    assert len(list(root.datasets())) == 4
    assert len(root.g1) == 7
    assert len(list(root.g1.groups())) == 3
    assert len(list(root.g1.datasets())) == 4
    assert len(root.g1.g2) == 5
    assert len(list(root.g1.g2.groups())) == 2
    assert len(list(root.g1.g2.datasets())) == 3
    assert len(root.g1.g2.g3) == 2
    assert len(list(root.g1.g2.g3.groups())) == 1
    assert len(list(root.g1.g2.g3.datasets())) == 1
    assert len(root.g1.g2.g3.g4) == 1
    assert len(list(root.g1.g2.g3.g4.groups())) == 0
    assert len(list(root.g1.g2.g3.g4.datasets())) == 1

    d1_removed = root.g1.g2.remove('d1')
    assert d1_removed is d1
    assert 'd1' not in root['/g1/g2']
    assert 'd1' not in root.g1['/g2']
    assert 'd1' not in root.g1.g2
    assert len(root) == 7
    assert len(list(root.groups())) == 4
    assert len(list(root.datasets())) == 3
    assert len(root.g1) == 6
    assert len(list(root.g1.groups())) == 3
    assert len(list(root.g1.datasets())) == 3
    assert len(root.g1.g2) == 4
    assert len(list(root.g1.g2.groups())) == 2
    assert len(list(root.g1.g2.datasets())) == 2
    assert len(root.g1.g2.g3) == 2
    assert len(list(root.g1.g2.g3.groups())) == 1
    assert len(list(root.g1.g2.g3.datasets())) == 1
    assert len(root.g1.g2.g3.g4) == 1
    assert len(list(root.g1.g2.g3.g4.groups())) == 0
    assert len(list(root.g1.g2.g3.g4.datasets())) == 1

    g3 = root.g1.g2.g3
    g3_removed = root.remove('/g1/g2/g3')
    assert g3 is g3_removed
    assert g3 not in root['/g1/g2']
    assert g3 not in root.g1['g2']
    assert len(root) == 4
    assert len(list(root.groups())) == 2
    assert len(list(root.datasets())) == 2
    assert len(root.g1) == 3
    assert len(list(root.g1.groups())) == 1
    assert len(list(root.g1.datasets())) == 2
    assert len(root.g1.g2) == 1
    assert len(list(root.g1.g2.groups())) == 0
    assert len(list(root.g1.g2.datasets())) == 1

    g2 = root['/g1/g2']
    g2_removed = root.g1.remove('g2')
    assert g2 is g2_removed
    assert g2 not in root['/g1']
    assert g2 not in root.g1
    assert len(root) == 2
    assert len(list(root.groups())) == 1
    assert len(list(root.datasets())) == 1
    assert len(root.g1) == 1
    assert len(list(root.g1.groups())) == 0
    assert len(list(root.g1.datasets())) == 1

    assert root.remove('a/b/c/d') is None


def test_get_ancestors():
    root = Root('')

    assert isinstance(root.ancestors(), types.GeneratorType)
    assert len(tuple(root.ancestors())) == 0

    root.create_group('a/b/c/d/e/f/g/h')
    root.a.create_dataset('dataset')

    x = root.create_group('x')
    x.require_group('y/z')
    root.x.y.create_dataset('dataset')

    ancestors = tuple(root.a.b.c.d.ancestors())
    assert len(ancestors) == 4
    assert ancestors[0] is root.a.b.c
    assert ancestors[1] is root.a.b
    assert ancestors[2] is root.a
    assert ancestors[3] is root

    h = root.a.b.c.d.e.f.g.h
    ancestors = tuple(h.ancestors())
    assert len(ancestors) == 8
    assert ancestors[0] is root.a.b.c.d.e.f.g
    assert ancestors[1] is root.a.b.c.d.e.f
    assert ancestors[2] is root.a.b.c.d.e
    assert ancestors[3] is root.a.b.c.d
    assert ancestors[4] is root.a.b.c
    assert ancestors[5] is root.a.b
    assert ancestors[6] is root.a
    assert ancestors[7] is root

    ancestors = tuple(root.x.y.z.ancestors())
    assert len(ancestors) == 3
    assert ancestors[0] is root.x.y
    assert ancestors[1] is root.x
    assert ancestors[2] is root

    ancestors = tuple(root.x.ancestors())
    assert len(ancestors) == 1
    assert ancestors[0] is root


def test_remove_same_endswith():
    def create_new_root():
        r = Root('')
        r.create_dataset('dset')
        a = r.create_group('a')
        a.create_dataset('dset')
        b = r.create_group('b')
        b.create_dataset('dset')
        c = b.create_group('c')
        c.create_dataset('dset')
        c.create_group('b')
        d = r.create_group('d')
        d.create_dataset('dset')
        d.create_group('c')
        d.c.create_dataset('dset')
        e = c.create_group('e')
        e.create_dataset('dset')
        return r

    root = create_new_root()
    assert len(list(root.groups())) == 7
    assert len(list(root.datasets())) == 7
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.b.groups())) == 3
    assert len(list(root.b.datasets())) == 3
    assert len(list(root.b.c.groups())) == 2
    assert len(list(root.b.c.datasets())) == 2
    assert len(list(root.b.c.e.groups())) == 0
    assert len(list(root.b.c.e.datasets())) == 1
    assert len(list(root.d.groups())) == 1
    assert len(list(root.d.datasets())) == 2
    assert len(list(root.d.c.groups())) == 0
    assert len(list(root.d.c.datasets())) == 1
    assert 'dset' in root
    assert 'a' in root
    assert 'dset' in root.a
    assert 'b' in root
    assert 'dset' in root.b
    assert 'c' in root.b
    assert 'dset' in root.b.c
    assert 'b' in root.b.c
    assert 'd' in root
    assert 'dset' in root.d
    assert 'c' in root.d
    assert 'dset' in root.d.c
    assert 'e' in root.b.c
    assert 'dset' in root.b.c.e

    ref = root['/b/c/e/dset']
    obj = root.remove('/b/c/e/dset')
    assert obj is ref
    assert len(list(root.groups())) == 7
    assert len(list(root.datasets())) == 6
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.b.groups())) == 3
    assert len(list(root.b.datasets())) == 2
    assert len(list(root.b.c.groups())) == 2
    assert len(list(root.b.c.datasets())) == 1
    assert len(list(root.b.c.e.groups())) == 0
    assert len(list(root.b.c.e.datasets())) == 0
    assert len(list(root.d.groups())) == 1
    assert len(list(root.d.datasets())) == 2
    assert len(list(root.d.c.groups())) == 0
    assert len(list(root.d.c.datasets())) == 1
    assert 'dset' in root  # does not get removed
    assert 'a' in root
    assert 'dset' in root.a  # does not get removed
    assert 'b' in root
    assert 'dset' in root.b  # does not get removed
    assert 'c' in root.b
    assert 'dset' in root.b.c  # does not get removed
    assert 'b' in root.b.c
    assert 'd' in root
    assert 'dset' in root.d  # does not get removed
    assert 'c' in root.d
    assert 'dset' in root.d.c  # does not get removed
    assert 'e' in root.b.c
    assert 'dset' not in root.b.c.e

    root = create_new_root()
    c = root.b.c
    ref = c.dset
    obj = c.remove('/dset')
    assert obj is ref
    assert len(list(root.groups())) == 7
    assert len(list(root.datasets())) == 6
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.b.groups())) == 3
    assert len(list(root.b.datasets())) == 2
    assert len(list(root.b.c.groups())) == 2
    assert len(list(root.b.c.datasets())) == 1
    assert len(list(root.b.c.e.groups())) == 0
    assert len(list(root.b.c.e.datasets())) == 1
    assert len(list(root.d.groups())) == 1
    assert len(list(root.d.datasets())) == 2
    assert len(list(root.d.c.groups())) == 0
    assert len(list(root.d.c.datasets())) == 1
    assert 'dset' in root  # does not get removed
    assert 'a' in root
    assert 'dset' in root.a  # does not get removed
    assert 'b' in root
    assert 'dset' in root.b  # does not get removed
    assert 'c' in root.b
    assert 'dset' not in root.b.c
    assert 'b' in root.b.c
    assert 'd' in root
    assert 'dset' in root.d  # does not get removed
    assert 'c' in root.d
    assert 'dset' in root.d.c  # does not get removed
    assert 'e' in root.b.c
    assert 'dset' in root.b.c.e  # does not get removed

    root = create_new_root()
    b = root.b
    ref = b.c.dset
    obj = b.remove('/c/dset')
    assert obj is ref
    assert len(list(root.groups())) == 7
    assert len(list(root.datasets())) == 6
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.b.groups())) == 3
    assert len(list(root.b.datasets())) == 2
    assert len(list(root.b.c.groups())) == 2
    assert len(list(root.b.c.datasets())) == 1
    assert len(list(root.b.c.e.groups())) == 0
    assert len(list(root.b.c.e.datasets())) == 1
    assert len(list(root.d.groups())) == 1
    assert len(list(root.d.datasets())) == 2
    assert len(list(root.d.c.groups())) == 0
    assert len(list(root.d.c.datasets())) == 1
    assert 'dset' in root  # does not get removed
    assert 'a' in root
    assert 'dset' in root.a  # does not get removed
    assert 'b' in root
    assert 'dset' in root.b  # does not get removed
    assert 'c' in root.b
    assert 'dset' not in root.b.c
    assert 'b' in root.b.c
    assert 'd' in root
    assert 'dset' in root.d  # does not get removed
    assert 'c' in root.d
    assert 'dset' in root.d.c  # does not get removed
    assert 'e' in root.b.c
    assert 'dset' in root.b.c.e  # does not get removed

    root = create_new_root()
    ref = root.b.c.b
    obj = root.b.c.remove('/b')
    assert obj is ref
    assert len(list(root.groups())) == 6
    assert len(list(root.datasets())) == 7
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.b.groups())) == 2
    assert len(list(root.b.datasets())) == 3
    assert len(list(root.b.c.groups())) == 1
    assert len(list(root.b.c.datasets())) == 2
    assert len(list(root.b.c.e.groups())) == 0
    assert len(list(root.b.c.e.datasets())) == 1
    assert len(list(root.d.groups())) == 1
    assert len(list(root.d.datasets())) == 2
    assert len(list(root.d.c.groups())) == 0
    assert len(list(root.d.c.datasets())) == 1
    assert 'dset' in root
    assert 'a' in root
    assert 'dset' in root.a
    assert 'b' in root  # does not get removed
    assert 'dset' in root.b
    assert 'c' in root.b
    assert 'dset' in root.b.c
    assert 'b' not in root.b.c
    assert 'd' in root
    assert 'dset' in root.d
    assert 'c' in root.d
    assert 'dset' in root.d.c
    assert 'e' in root.b.c
    assert 'dset' in root.b.c.e

    root = create_new_root()
    ref = root.b.c.b
    obj = root.remove('/b/c/b')
    assert obj is ref
    assert len(list(root.groups())) == 6
    assert len(list(root.datasets())) == 7
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.b.groups())) == 2
    assert len(list(root.b.datasets())) == 3
    assert len(list(root.b.c.groups())) == 1
    assert len(list(root.b.c.datasets())) == 2
    assert len(list(root.b.c.e.groups())) == 0
    assert len(list(root.b.c.e.datasets())) == 1
    assert len(list(root.d.groups())) == 1
    assert len(list(root.d.datasets())) == 2
    assert len(list(root.d.c.groups())) == 0
    assert len(list(root.d.c.datasets())) == 1
    assert 'dset' in root
    assert 'a' in root
    assert 'dset' in root.a
    assert 'b' in root  # does not get removed
    assert 'dset' in root.b
    assert 'c' in root.b
    assert 'dset' in root.b.c
    assert 'b' not in root.b.c
    assert 'd' in root
    assert 'dset' in root.d
    assert 'c' in root.d
    assert 'dset' in root.d.c
    assert 'e' in root.b.c
    assert 'dset' in root.b.c.e

    root = create_new_root()
    ref = root.b
    obj = root.remove('/b')
    assert obj is ref
    assert len(list(root.groups())) == 3
    assert len(list(root.datasets())) == 4
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.d.groups())) == 1
    assert len(list(root.d.datasets())) == 2
    assert len(list(root.d.c.groups())) == 0
    assert len(list(root.d.c.datasets())) == 1
    assert 'dset' in root
    assert 'a' in root
    assert 'dset' in root.a
    assert 'b' not in root
    assert 'd' in root
    assert 'dset' in root.d
    assert 'c' in root.d
    assert 'dset' in root.d.c

    root = create_new_root()
    b = root.b
    ref = root.b.c.b
    obj = b.remove('/c/b')
    assert obj is ref
    assert len(list(root.groups())) == 6
    assert len(list(root.datasets())) == 7
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.b.groups())) == 2
    assert len(list(root.b.datasets())) == 3
    assert len(list(root.b.c.groups())) == 1
    assert len(list(root.b.c.datasets())) == 2
    assert len(list(root.b.c.e.groups())) == 0
    assert len(list(root.b.c.e.datasets())) == 1
    assert len(list(root.d.groups())) == 1
    assert len(list(root.d.datasets())) == 2
    assert len(list(root.d.c.groups())) == 0
    assert len(list(root.d.c.datasets())) == 1
    assert 'dset' in root
    assert 'a' in root
    assert 'dset' in root.a
    assert 'b' in root
    assert 'dset' in root.b
    assert 'c' in root.b
    assert 'dset' in root.b.c
    assert 'b' not in root.b.c
    assert 'd' in root
    assert 'dset' in root.d
    assert 'c' in root.d
    assert 'dset' in root.d.c
    assert 'e' in root.b.c
    assert 'dset' in root.b.c.e

    root = create_new_root()
    ref = root.b.c.b
    obj = root.remove('/b/c/b')
    assert obj is ref
    assert len(list(root.groups())) == 6
    assert len(list(root.datasets())) == 7
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.b.groups())) == 2
    assert len(list(root.b.datasets())) == 3
    assert len(list(root.b.c.groups())) == 1
    assert len(list(root.b.c.datasets())) == 2
    assert len(list(root.b.c.e.groups())) == 0
    assert len(list(root.b.c.e.datasets())) == 1
    assert len(list(root.d.groups())) == 1
    assert len(list(root.d.datasets())) == 2
    assert len(list(root.d.c.groups())) == 0
    assert len(list(root.d.c.datasets())) == 1
    assert 'dset' in root
    assert 'a' in root
    assert 'dset' in root.a
    assert 'b' in root
    assert 'dset' in root.b
    assert 'c' in root.b
    assert 'dset' in root.b.c
    assert 'b' not in root.b.c
    assert 'd' in root
    assert 'dset' in root.d
    assert 'c' in root.d
    assert 'dset' in root.d.c
    assert 'e' in root.b.c
    assert 'dset' in root.b.c.e

    root = create_new_root()
    c = root.b.c
    ref = c.e
    obj = c.remove('e')
    assert obj is ref
    assert len(list(root.groups())) == 6
    assert len(list(root.datasets())) == 6
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.b.groups())) == 2
    assert len(list(root.b.datasets())) == 2
    assert len(list(root.b.c.groups())) == 1
    assert len(list(root.b.c.datasets())) == 1
    assert len(list(root.d.groups())) == 1
    assert len(list(root.d.datasets())) == 2
    assert len(list(root.d.c.groups())) == 0
    assert len(list(root.d.c.datasets())) == 1
    assert 'dset' in root
    assert 'a' in root
    assert 'dset' in root.a
    assert 'b' in root
    assert 'dset' in root.b
    assert 'c' in root.b
    assert 'dset' in root.b.c
    assert 'b' in root.b.c
    assert 'd' in root
    assert 'dset' in root.d
    assert 'c' in root.d
    assert 'dset' in root.d.c
    assert 'e' not in root.b.c

    root = create_new_root()
    assert root.remove('invalid') is None

    root = create_new_root()
    ref = root.d.c
    obj = root.remove('/d/c/')
    assert obj is ref
    assert len(list(root.groups())) == 6
    assert len(list(root.datasets())) == 6
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.b.groups())) == 3
    assert len(list(root.b.datasets())) == 3
    assert len(list(root.b.c.groups())) == 2
    assert len(list(root.b.c.datasets())) == 2
    assert len(list(root.b.c.e.groups())) == 0
    assert len(list(root.b.c.e.datasets())) == 1
    assert len(list(root.d.groups())) == 0
    assert len(list(root.d.datasets())) == 1
    assert 'dset' in root
    assert 'a' in root
    assert 'dset' in root.a
    assert 'b' in root
    assert 'dset' in root.b
    assert 'c' in root.b
    assert 'dset' in root.b.c
    assert 'b' in root.b.c
    assert 'd' in root
    assert 'dset' in root.d
    assert 'c' not in root.d
    assert 'e' in root.b.c
    assert 'dset' in root.b.c.e

    root = create_new_root()
    ref = root.d.c
    obj = root.d.remove('c')
    assert obj is ref
    assert len(list(root.groups())) == 6
    assert len(list(root.datasets())) == 6
    assert len(list(root.a.groups())) == 0
    assert len(list(root.a.datasets())) == 1
    assert len(list(root.b.groups())) == 3
    assert len(list(root.b.datasets())) == 3
    assert len(list(root.b.c.groups())) == 2
    assert len(list(root.b.c.datasets())) == 2
    assert len(list(root.b.c.e.groups())) == 0
    assert len(list(root.b.c.e.datasets())) == 1
    assert len(list(root.d.groups())) == 0
    assert len(list(root.d.datasets())) == 1
    assert 'dset' in root
    assert 'a' in root
    assert 'dset' in root.a
    assert 'b' in root
    assert 'dset' in root.b
    assert 'c' in root.b
    assert 'dset' in root.b.c
    assert 'b' in root.b.c
    assert 'd' in root
    assert 'dset' in root.d
    assert 'c' not in root.d
    assert 'e' in root.b.c
    assert 'dset' in root.b.c.e



from datetime import datetime

import pytest
from googleapiclient.errors import HttpError

from msl.io import GDrive
from msl.io import GSheetsReader
from msl.io.readers.gsheets import _google_file_id_regex

try:
    # dr -> drive, readonly
    dr = GDrive(account='testing', read_only=True)
except:
    dr = None

try:
    # sr -> sheets, readonly
    sr = GSheetsReader('1TI3pM-534SZ5DQTEZ-7vCI04l48f8ZpLGbfEWJuCFSo', account='testing', read_only=True)
except:
    sr = None

skipif_no_gdrive_readonly = pytest.mark.skipif(
    dr is None, reason='No GDrive readonly token'
)

skipif_no_sheets_readonly = pytest.mark.skipif(
    sr is None, reason='No GSheets readonly token'
)


@skipif_no_sheets_readonly
def test_raises():
    with pytest.raises(ValueError, match=r'Must instantiate GSheetsReader in read-only mode'):
        GSheetsReader('does-not-matter', read_only=False)

    table_gsheet_id = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ'
    with pytest.raises(ValueError, match=r'You must specify the name of the sheet to read'):
        GSheetsReader(table_gsheet_id, account='testing').read()

    table_gsheet_id = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ'
    with pytest.raises(ValueError, match=r'There is no sheet named'):
        GSheetsReader(table_gsheet_id, account='testing').read(sheet='A1')

    table_gsheet_id = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ'
    with pytest.raises(HttpError):
        GSheetsReader(table_gsheet_id, account='testing').read(sheet='SheetA1')


@skipif_no_sheets_readonly
def test_cell():
    # lab_environment.gsheet
    ssid = '1TI3pM-534SZ5DQTEZ-7vCI04l48f8ZpLGbfEWJuCFSo'
    values = [('temperature', 'humidity'), (20.33, 49.82), (20.23, 46.06), (20.41, 47.06), (20.29, 48.32)]

    sheets = GSheetsReader(ssid, account='testing', read_only=True)
    assert sheets.file == ssid
    assert sheets.read() == values
    assert sheets.read(cell='A1') == 'temperature'
    assert sheets.read(cell='A100') is None  # A100 is empty
    assert sheets.read(cell='B5') == 48.32
    assert sheets.read(cell='A3:B3') == [(20.23, 46.06)]
    assert sheets.read(cell='A6:B6') == []  # row 6 is empty
    assert sheets.read(cell='B:B') == [('humidity',), (49.82,), (46.06,), (47.06,), (48.32,)]
    assert sheets.read(cell='C:C') == []  # column C is empty
    assert sheets.read(cell='A:B') == values
    assert sheets.read(cell='A1:B5') == values
    assert sheets.read(cell='A1:Z100') == values  # slicing out of range is okay
    assert sheets.read(cell='J1:M10') == []


@skipif_no_sheets_readonly
def test_sheet():
    # table.gsheet
    ssid = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ'
    sheets = GSheetsReader(ssid, account='testing', read_only=True)
    assert sheets.file == ssid
    assert sheets.sheet_names() == ('StartA1', 'StartH22', 'header only', 'empty', 'column', 'row')
    assert sheets.read(sheet='empty') == []
    assert sheets.read(sheet='header only') == [('Timestamp', 'Value', 'Valid', 'ID')]
    assert sheets.read(sheet='column') == [
        ('Value',), (20.1,), (25.4,), (19.4,), (11.8,), (24.6,), (20.7,), (21.8,), (19.2,), (18.6,), (16.4,)
    ]
    assert sheets.read(sheet='row') == [
        ('Timestamp', 'Value', 'Valid', 'ID'),
        (datetime(2019, 9, 11, 14, 6, 55), 20.1, True, 'sensor 1')
    ]

    assert sheets.read('A1:A1', sheet='StartA1') == [('Timestamp',)]
    assert sheets.read('D9:D9', sheet='StartA1') == [('sensor 3',)]


@skipif_no_sheets_readonly
def test_as_datetime():
    # table.gsheet
    ssid = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ'
    sheets = GSheetsReader(ssid, account='testing', read_only=True)
    assert sheets.read(cell='A:A', sheet='StartA1', as_datetime=True) == [
        ('Timestamp',),
        (datetime(2019, 9, 11, 14, 6, 55),),
        (datetime(2019, 9, 11, 14, 6, 59),),
        (datetime(2019, 9, 11, 14, 7, 3),),
        (datetime(2019, 9, 11, 14, 7, 7),),
        (datetime(2019, 9, 11, 14, 7, 11),),
        (datetime(2019, 9, 11, 14, 7, 15),),
        (datetime(2019, 9, 11, 14, 7, 19),),
        (datetime(2019, 9, 11, 14, 7, 23),),
        (datetime(2019, 9, 11, 14, 7, 27),),
        (datetime(2019, 9, 11, 14, 7, 31),)
    ]
    assert sheets.read(cell='A:B', sheet='StartA1', as_datetime=False) == [
        ('Timestamp', 'Value'),
        ('2019-09-11 14:06:55', 20.1),
        ('2019-09-11 14:06:59', 25.4),
        ('2019-09-11 14:07:03', 19.4),
        ('2019-09-11 14:07:07', 11.8),
        ('2019-09-11 14:07:11', 24.6),
        ('2019-09-11 14:07:15', 20.7),
        ('2019-09-11 14:07:19', 21.8),
        ('2019-09-11 14:07:23', 19.2),
        ('2019-09-11 14:07:27', 18.6),
        ('2019-09-11 14:07:31', 16.4),
    ]


@skipif_no_gdrive_readonly
@skipif_no_sheets_readonly
def test_file_path():
    path = 'My Drive/MSL/msl-io-testing/empty-5.gsheet'
    sheets = GSheetsReader(path, account='testing', read_only=True)
    assert sheets.file == path
    assert sheets.sheet_names() == ('Sheet1', 'Sheet2', 'Sheet3', 'Sheet4', 'Sheet5')
    assert sheets.read(sheet='Sheet1') == []

    sheets = GSheetsReader('table', account='testing', read_only=True)
    assert sheets.file == 'table'
    assert sheets.read(sheet='empty') == []


def test_google_file_id_regex():
    assert _google_file_id_regex.search('1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ')
    assert _google_file_id_regex.search('1IemLij3ggB_S5ASO7qyPSIQUmvhWgBfemePn7guAJe4')
    assert _google_file_id_regex.search('1IemLij3ggB-S5ASO7qyPSIQUmvhWgBfemePn7guAJe4')
    assert _google_file_id_regex.search('1IemL_j3ggB-S5ASO-qyPSIQUmv_WgBfe-ePn7gu-Je4')

    # does not start with 1
    assert not _google_file_id_regex.search('Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ1')

    # not 44 characters
    for n in [0, 10, 40, 42, 43, 45, 50, 100]:
        assert not _google_file_id_regex.search('1' * n)

    # contains an invalid character
    for c in r""" !"#$%&'()*+,./:;<=>?@[\]^`{|}~""":
        assert not _google_file_id_regex.search('1Q0TAgnw6AJQWkLMf{}V3qEhEXuCEXTFAc95cEcshOXnQ'.format(c))



import os
import pathlib
from io import BytesIO
from io import StringIO

import pytest

from msl.io import Reader
from msl.io.base import Root
from msl.io.constants import IS_PYTHON2


def test_get_root():
    root = Reader('')
    assert isinstance(root, Root)
    assert not root.read_only


def test_instantiate():
    reader = Reader('aaa.bbb')
    assert reader.file == 'aaa.bbb'


def test_get_lines():
    path = os.path.join(os.path.dirname(__file__), 'samples', 'test_file_for_static_Reader_lines')

    # the file contains 26 lines
    with open(path, mode='rt') as fp:
        all_lines = fp.read().split('\n')

    string_io = StringIO()
    with open(path, mode='rt') as fp:
        data = fp.read()
        if IS_PYTHON2:
            data = unicode(data)
        string_io.write(data)
    string_io.seek(0)

    open_ = open(path, mode='rt')

    for obj in [path, string_io, open_]:
        assert len(Reader.get_lines(obj)) == 26
        assert len(Reader.get_lines(obj, remove_empty_lines=True)) == 24

        assert Reader.get_lines(obj) == all_lines
        assert Reader.get_lines(obj, None) == all_lines
        assert Reader.get_lines(obj, 0) == []
        assert Reader.get_lines(obj, 1) == ['line1']
        assert Reader.get_lines(obj, -1) == ['line26']
        assert Reader.get_lines(obj, 5) == ['line1', 'line2', 'line3', 'line4', 'line5']
        assert Reader.get_lines(obj, -5) == ['line22', 'line23', 'line24', 'line25', 'line26']
        assert Reader.get_lines(obj, 100) == all_lines
        assert Reader.get_lines(obj, -100) == all_lines

        assert Reader.get_lines(obj, None, None) == all_lines
        assert Reader.get_lines(obj, None, 0) == []
        assert Reader.get_lines(obj, None, 1) == ['line1']
        assert Reader.get_lines(obj, None, -1) == all_lines
        assert Reader.get_lines(obj, None, 5) == ['line1', 'line2', 'line3', 'line4', 'line5']
        assert Reader.get_lines(obj, None, -20) == ['line1', 'line2', 'line3', 'line4', 'line5', 'line6', 'line7']
        assert Reader.get_lines(obj, None, 100) == all_lines
        assert Reader.get_lines(obj, None, -100) == []

        assert Reader.get_lines(obj, 0, None) == all_lines
        assert Reader.get_lines(obj, 1, None) == all_lines
        assert Reader.get_lines(obj, -1, None) == ['line26']
        assert Reader.get_lines(obj, 18, None) == ['line18', 'line19', 'line20', '', 'line22',
                                                   'line23', 'line24', 'line25', 'line26']
        assert Reader.get_lines(obj, -5, None) == ['line22', 'line23', 'line24', 'line25', 'line26']
        assert Reader.get_lines(obj, 100, None) == []  # there are only 26 lines
        assert Reader.get_lines(obj, -100, None) == all_lines

        assert Reader.get_lines(obj, 0, 0) == []
        assert Reader.get_lines(obj, 1, 1) == ['line1']
        assert Reader.get_lines(obj, 1, -1) == all_lines
        assert Reader.get_lines(obj, 4, 8) == ['line4', 'line5', 'line6', 'line7', 'line8']
        assert Reader.get_lines(obj, -8, -4) == ['line19', 'line20', '', 'line22', 'line23']
        assert Reader.get_lines(obj, 2, 4) == ['line2', 'line3', 'line4']
        assert Reader.get_lines(obj, -5, 4) == []
        assert Reader.get_lines(obj, 10, -7) == ['line10', '', 'line12', 'line13', 'line14', 'line15',
                                                 'line16', 'line17', 'line18', 'line19', 'line20']
        assert Reader.get_lines(obj, 100, 200) == []  # there are only 26 lines
        assert Reader.get_lines(obj, -100, -50) == []
        assert Reader.get_lines(obj, 25, 100) == ['line25', 'line26']

        assert Reader.get_lines(obj, 1, -1, 6) == ['line1', 'line7', 'line13', 'line19', 'line25']
        assert Reader.get_lines(obj, 0, None, 6) == ['line1', 'line7', 'line13', 'line19', 'line25']
        assert Reader.get_lines(obj, None, None, 6) == ['line1', 'line7', 'line13', 'line19', 'line25']
        assert Reader.get_lines(obj, 1, 15, 6) == ['line1', 'line7', 'line13']
        assert Reader.get_lines(obj, -20, -5, 5) == ['line7', 'line12', 'line17', 'line22']
        assert Reader.get_lines(obj, -100, -21, 2) == ['line1', 'line3', 'line5']
        assert Reader.get_lines(obj, -100, -20, 2) == ['line1', 'line3', 'line5', 'line7']
        assert Reader.get_lines(obj, 15, 25, 3) == ['line15', 'line18', '', 'line24']
        assert Reader.get_lines(obj, 15, 25, 3, remove_empty_lines=True) == ['line15', 'line18', 'line24']

    string_io.close()
    open_.close()


def test_get_bytes():
    path = os.path.join(os.path.dirname(__file__), 'samples', 'test_file_for_static_Reader_bytes')

    # the file contains 184 bytes
    with open(path, mode='rb') as fp:
        all_bytes = fp.read()

    bytes_io = BytesIO()
    with open(path, mode='rb') as fp:
        bytes_io.write(fp.read())
    bytes_io.seek(0)

    open_ = open(path, mode='rb')

    for obj in [path, bytes_io, open_]:
        assert Reader.get_bytes(obj) == all_bytes
        assert Reader.get_bytes(obj, None) == all_bytes
        assert Reader.get_bytes(obj, 0) == b''
        assert Reader.get_bytes(obj, 1) == b'!'
        assert Reader.get_bytes(obj, -1) == b'~'
        assert Reader.get_bytes(obj, 7) == b'!"#$%&('
        assert Reader.get_bytes(obj, -5) == b'z{|}~'
        assert Reader.get_bytes(obj, -21) == b'jklmnopqrstuvwxyz{|}~'
        assert Reader.get_bytes(obj, -5000) == all_bytes
        assert Reader.get_bytes(obj, 5000) == all_bytes

        assert Reader.get_bytes(obj, None, None) == all_bytes
        assert Reader.get_bytes(obj, None, 0) == b''
        assert Reader.get_bytes(obj, None, -1) == all_bytes
        assert Reader.get_bytes(obj, None, 1) == b'!'
        assert Reader.get_bytes(obj, None, -179) == b'!"#$%&'  # 184 - 179 -> the first 6 bytes
        assert Reader.get_bytes(obj, None, 8) == b'!"#$%&()'
        assert Reader.get_bytes(obj, None, -5000) == b''
        assert Reader.get_bytes(obj, None, 5000) == all_bytes

        assert Reader.get_bytes(obj, 0, None) == all_bytes
        assert Reader.get_bytes(obj, 1, None) == all_bytes
        assert Reader.get_bytes(obj, -1, None) == b'~'
        assert Reader.get_bytes(obj, 123, None) == b'@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_`abcdefghijklmnopqrstuvwxyz{|}~'
        assert Reader.get_bytes(obj, -37, None) == b'YZ[]^_`abcdefghijklmnopqrstuvwxyz{|}~'
        assert Reader.get_bytes(obj, -5000, None) == all_bytes
        assert Reader.get_bytes(obj, 5000, None) == b''

        assert Reader.get_bytes(obj, 0, 0) == b''
        assert Reader.get_bytes(obj, 1, 1) == b'!'
        assert Reader.get_bytes(obj, 1, -1) == all_bytes
        assert Reader.get_bytes(obj, 5, 10) == b'%&()*+'
        assert Reader.get_bytes(obj, 139, -1) == b'PQRSTUVWXYZ[]^_`abcdefghijklmnopqrstuvwxyz{|}~'
        assert Reader.get_bytes(obj, 123, -20) == b'@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_`abcdefghijk'
        assert Reader.get_bytes(obj, -101, 55) == b''
        assert Reader.get_bytes(obj, 33, 57) == b'BCDEFGHIJKLMNOPQRSTUVWXYZ'
        assert Reader.get_bytes(obj, -10, -4) == b'uvwxyz{'
        assert Reader.get_bytes(obj, 600, -600) == b''
        assert Reader.get_bytes(obj, 100, 50) == b''
        assert Reader.get_bytes(obj, 5000, 6000) == b''
        assert Reader.get_bytes(obj, -6000, -5000) == b''

        assert Reader.get_bytes(obj, 0, 6, 3) == b'!$'
        assert Reader.get_bytes(obj, 1, 6, 3) == b'!$'
        assert Reader.get_bytes(obj, 0, 7, 3) == b'!$('
        assert Reader.get_bytes(obj, 1, 7, 3) == b'!$('
        assert Reader.get_bytes(obj, 0, 8, 3) == b'!$('
        assert Reader.get_bytes(obj, 1, 8, 3) == b'!$('
        assert Reader.get_bytes(obj, 0, 12, 3) == b'!$(+'
        assert Reader.get_bytes(obj, 1, 12, 3) == b'!$(+'
        assert Reader.get_bytes(obj, 0, 13, 3) == b'!$(+.'
        assert Reader.get_bytes(obj, 1, 13, 3) == b'!$(+.'
        assert Reader.get_bytes(obj, 9, 49, 8) == b'*2:BJR'
        assert Reader.get_bytes(obj, 9, 53, 8) == b'*2:BJR'
        assert Reader.get_bytes(obj, -19, -5, 5) == b'lqv'
        assert Reader.get_bytes(obj, -19, -4, 5) == b'lqv{'
        assert Reader.get_bytes(obj, -10, -1, 2) == b'uwy{}'
        assert Reader.get_bytes(obj, -11, -1, 2) == b'tvxz|~'
        assert Reader.get_bytes(obj, -200, -155, 5) == b'!&,16;'
        assert Reader.get_bytes(obj, 109, 500, 10) == b'2<FPZeoy'

    bytes_io.close()
    open_.close()


def test_get_extension():
    assert Reader.get_extension('') == ''
    assert Reader.get_extension('xxx') == ''
    assert Reader.get_extension('a.xxx') == '.xxx'
    assert Reader.get_extension('/home/msl/data.csv') == '.csv'
    assert Reader.get_extension('/home/msl/filename.with.dots.dat') == '.dat'
    assert Reader.get_extension(StringIO()) == ''
    assert Reader.get_extension(BytesIO()) == ''
    assert Reader.get_extension(pathlib.Path('')) == ''
    assert Reader.get_extension(pathlib.Path('a.x')) == '.x'
    assert Reader.get_extension(pathlib.Path(r'C:\folder\hello.world')) == '.world'
    assert Reader.get_extension(pathlib.Path('filename.with.dots.dat')) == '.dat'

    path = os.path.join(os.path.dirname(__file__), 'samples', 'excel_datatypes.xlsx')
    with open(path, mode='rt') as fp:
        assert Reader.get_extension(fp) == '.xlsx'

    path = os.path.join(os.path.dirname(__file__), 'samples', 'test_file_for_static_Reader_lines')
    with open(path, mode='rt') as fp:
        assert Reader.get_extension(fp) == ''


def test_override_methods():

    # the subclass must override this method
    with pytest.raises(NotImplementedError):
        Reader('').read()

    # the subclass must override this method
    assert not Reader('').can_read('')



"""
Writer for a JSON_ file format. The corresponding :class:`~msl.io.base.Reader` is
:class:`~msl.io.readers.json_.JSONReader`.

.. _JSON: https://www.json.org/
"""
import codecs
import json
import os
from io import BufferedIOBase

import numpy as np

from ..base import Root
from ..base import Writer
from ..constants import IS_PYTHON2
from ..metadata import Metadata
from ..utils import is_file_readable

# Custom JSON encoder that writes a 1-dimensional list on a single line.
if IS_PYTHON2:
    from ._py2_json_encoder import _make_iterencode
else:
    from ._py3_json_encoder import _make_iterencode

_original_make_iterencode = json.encoder._make_iterencode


class JSONWriter(Writer):
    """Create a JSON_ writer.

    You can use :class:`JSONWriter` as a :ref:`context manager <with>`.
    For example,

    .. code-block:: pycon

        >>> with JSONWriter('example.json') as root:
        ...     dset = root.create_dataset('dset', data=[1, 2, 3])
        ...     root.update_context_kwargs(indent=4)

    .. invisible-code-block: pycon

        >>> import os
        >>> os.remove('example.json')

    This will automatically write `root` to the specified file using
    ``indent=4`` as a keyword argument to the :meth:`.write` method when
    the :ref:`with <with>` block exits.
    """

    def write(self, file=None, root=None, **kwargs):
        """Write to a JSON_ file.

        The first line in the output file contains a description that the
        file was created by the :class:`JSONWriter`. It begins with a ``#`` and
        contains a version number.

        Version 1.0 specifications

            * Use the ``'dtype'`` and ``'data'`` keys to uniquely identify a
              JSON_ object as a :class:`~msl.io.dataset.Dataset`.

            * If a :class:`~msl.io.metadata.Metadata` `key` has a `value` that is a
              :class:`~msl.io.metadata.Metadata` object then the `key` becomes the name
              of a :class:`~msl.io.group.Group` and the `value` becomes
              :class:`~msl.io.metadata.Metadata` of that :class:`~msl.io.group.Group`.

        Parameters
        ----------
        file : :term:`path-like <path-like object>` or :term:`file-like <file object>`, optional
            The file to write the `root` to. If :data:`None` then uses the value of
            `file` that was specified when :class:`JSONWriter` was instantiated.
        root : :class:`~msl.io.base.Root`, optional
            Write `root` in JSON_ format. If :data:`None` then write the
            :class:`~msl.io.group.Group`\\s and :class:`~msl.io.dataset.Dataset`\\s
            in this :class:`JSONWriter`.
        **kwargs
            Accepts `mode`, `encoding` and `errors` keyword arguments which are passed
            to :func:`open`. The default `encoding` value is ``'utf-8'`` and the default
            `errors` value is ``'strict'``. All additional keyword arguments are passed to
            `json.dump <https://docs.python.org/3/library/json.html#json.dump>`_.
            The default indentation is 2.
        """
        if file is None:
            file = self.file
        if not file:
            raise ValueError('You must specify a file to write the root to')

        if root is None:
            root = self
        elif not isinstance(root, Root):
            raise TypeError('The root parameter must be a Root object')

        def add_dataset(d, dataset):
            if dataset.dtype.fields:
                # can't iterate over dataset.dtype.fields.items() since Python < 3.6
                # does not preserve order in a dict
                fields = dataset.dtype.fields
                d['dtype'] = [[n, str(fields[n][0])] for n in dataset.dtype.names]
            else:
                d['dtype'] = dataset.dtype.str
            d['data'] = dataset.tolist()

        def meta_to_dict(metadata):
            return dict((k, meta_to_dict(v) if isinstance(v, Metadata) else v)
                        for k, v in metadata.items())

        dict_ = dict(**meta_to_dict(root.metadata))

        for name, value in root.items():
            vertices = name.split('/')
            root_key = vertices[1]

            if root_key not in dict_:
                dict_[root_key] = dict(**meta_to_dict(value.metadata))
                if root.is_dataset(value):
                    add_dataset(dict_[root_key], value)

            if len(vertices) > 2:
                vertex = dict_[root_key]
                for key in vertices[2:-1]:
                    vertex = vertex[key]

                leaf_key = vertices[-1]
                if leaf_key not in vertex:
                    vertex[leaf_key] = dict(**meta_to_dict(value.metadata))
                    if root.is_dataset(value):
                        add_dataset(vertex[leaf_key], value)

        open_kwargs = {
            'mode': kwargs.pop('mode', None),
            'encoding': kwargs.pop('encoding', 'utf-8'),
            'errors': kwargs.pop('errors', 'strict')
        }

        if IS_PYTHON2:
            opener = codecs.open  # this allows the encoding and errors kwargs to be used
        else:
            # don't use codecs.open because the file looks better when opened in Notepad
            # (on Windows) when the standard open function use used
            opener = open

        is_file_like = hasattr(file, 'write')

        if not is_file_like:
            if not open_kwargs['mode'] or (open_kwargs['mode'] == 'x' and IS_PYTHON2):
                open_kwargs['mode'] = 'w'
                if os.path.isfile(file) or is_file_readable(file):
                    raise OSError(
                        "File exists {!r}\n"
                        "Specify mode='w' if you want to overwrite it.".format(file)
                    )
            elif open_kwargs['mode'] == 'r':
                raise ValueError('Invalid mode {!r}'.format(open_kwargs['mode']))
            elif open_kwargs['mode'] == 'a':
                open_kwargs['mode'] = 'w'

        if 'indent' not in kwargs:
            kwargs['indent'] = 2
        if 'cls' not in kwargs:
            kwargs['cls'] = _NumpyEncoder
            json.encoder._make_iterencode = _make_iterencode

        # header = '#File created with: MSL {} version 1.0\n'.format(self.__class__.__name__)
        #
        # Don't use the above definition of 'header' since JSONWriter could be sub-classed
        # and therefore the value of self.__class__.__name__ would change. The
        # JSONReader.can_read() method expects the text 'MSL JSONWriter' to be in a
        # specific location on the first line in the file.
        header = '#File created with: MSL JSONWriter version 1.0\n'

        if is_file_like:
            if isinstance(file, BufferedIOBase):  # a bytes-like object is required
                encoding = open_kwargs['encoding']
                file.write(header.encode(encoding))
                file.write(json.dumps(dict_, **kwargs).encode(encoding))
            elif IS_PYTHON2:
                file.write(unicode(header))
                file.write(unicode(json.dumps(dict_, **kwargs)))
            else:
                file.write(header)
                json.dump(dict_, file, **kwargs)
        else:
            with opener(file, **open_kwargs) as fp:
                fp.write(header)
                json.dump(dict_, fp, **kwargs)

        if kwargs['cls'] is _NumpyEncoder:
            json.encoder._make_iterencode = _original_make_iterencode


class _NumpyEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, bytes):
            return obj.decode(encoding='utf-8')

        # Let the base class raise the TypeError
        return json.JSONEncoder.default(self, obj)



import os
import tempfile

import numpy as np
import pytest

try:
    import h5py
except ImportError:
    h5py = None

from helper import read_sample
from helper import roots_equal
from msl.io import HDF5Writer
from msl.io import JSONWriter
from msl.io import read
from msl.io.readers import HDF5Reader


@pytest.mark.skipif(h5py is None, reason='h5py not installed')
def test_read_write_convert():
    root1 = read_sample('hdf5_sample.h5')

    # write as HDF5 then read
    writer = HDF5Writer(tempfile.gettempdir() + '/msl-hdf5-writer-temp.h5')
    writer.write(root=root1, mode='w')
    root2 = read(writer.file)
    assert root2.file == writer.file
    assert roots_equal(root1, root2)
    os.remove(writer.file)

    # convert to JSON then back to HDF5
    json_writer = JSONWriter(tempfile.gettempdir() + '/msl-json-writer-temp.json')
    json_writer.write(root=root1, mode='w')
    root_json = read(json_writer.file)
    assert root_json.file == json_writer.file
    assert roots_equal(root1, root_json)
    os.remove(json_writer.file)
    writer2 = HDF5Writer(tempfile.gettempdir() + '/msl-hdf5-writer-temp2.h5')
    writer2.write(root=root_json, mode='w')
    root3 = read(writer2.file)
    assert root3.file == writer2.file
    assert roots_equal(root1, root3)
    os.remove(writer2.file)

    for root in [root1, root2, root3]:
        assert isinstance(root, HDF5Reader)

        for key, value in root.items():
            k, v = str(key), str(value)
            k, v = repr(key), repr(value)

        order = ['D0', 'G0', 'G1A', 'D1', 'G1B', 'D2', 'D3', 'G2']
        for i, key in enumerate(root.keys()):
            assert os.path.basename(key) == order[i]

        assert len(root.metadata) == 3
        assert root.metadata['version_h5py'] == '2.8.0'
        assert root.metadata.version_hdf5 == '1.10.2'
        assert root.metadata['date_created'] == '2018-08-28 15:16:43.904990'

        assert 'D0' in root
        assert 'G0' in root

        d0 = root['D0']
        assert root.is_dataset(d0)
        assert d0.shape == (10, 4)
        assert d0.dtype.str == '<f4'
        assert len(d0.metadata) == 2
        assert d0.metadata['temperature'] == 21.2
        assert d0.metadata.temperature_units == 'deg C'

        g0 = root.G0
        assert root.is_group(g0)
        assert len(g0.metadata) == 1
        assert all(g0.metadata['count'] == [1, 2, 3, 4, 5])
        assert 'G1A' in g0
        assert 'G1B' in g0

        g1a = g0['G1A']
        assert root.is_group(g1a)
        assert len(g1a.metadata) == 2
        assert g1a.metadata['one'] == 1
        assert g1a.metadata['a'] == 'A'

        g1b = g0['G1B']
        assert root.is_group(g1b)
        assert len(g1b.metadata) == 2
        assert g1b.metadata['one'] == 1
        assert g1b.metadata['b'] == 'B'

        assert 'D1' in g0['G1A']
        d1 = root.G0.G1A.D1
        assert root.is_dataset(d1)
        assert len(d1.metadata) == 0
        assert d1.shape == (3, 3)
        assert d1.dtype.str == '<f8'

        assert 'D2' in g1b
        assert 'D3' in g0.G1B
        assert 'G2' in root.G0.G1B

        d2 = g1b['D2']
        assert root.is_dataset(d2)
        assert len(d2.metadata) == 2
        assert d2.metadata['voltage'] == 132.4
        assert d2.metadata['voltage_units'] == 'uV'
        assert d2.shape == (10,)
        assert d2.dtype.str == '<i4'
        assert d2[3] == 90

        d3 = g1b.D3
        assert root.is_dataset(d3)
        assert len(d3.metadata) == 0
        assert d3.shape == (10,)
        assert d3.dtype.str == '<i4'
        assert d3[7] == 51

        g2 = root.G0.G1B.G2
        assert root.is_group(g2)
        assert len(g2.metadata) == 1
        assert g2.metadata['hello'] == 'world'


@pytest.mark.skipif(h5py is None, reason='h5py not installed')
def test_raises():
    root = read_sample('hdf5_sample.h5')

    writer = HDF5Writer()
    assert writer.file is None

    # no file was specified
    with pytest.raises(ValueError, match=r'must specify a file'):
        writer.write(root=root)

    # root must be a Root object
    with pytest.raises(TypeError, match=r'Root'):
        writer.write(file='whatever', root=list(root.datasets())[0])
    with pytest.raises(TypeError, match=r'Root'):
        writer.write(file='whatever', root=list(root.groups())[0])
    with pytest.raises(TypeError, match=r'Root'):
        writer.write(file='whatever', root='Root')

    # cannot overwrite a file by default
    file = tempfile.gettempdir() + '/msl-hdf5-writer-temp.h5'
    with open(file, mode='wt') as fp:
        fp.write('Hi')
    with pytest.raises(OSError, match=r'File exists'):
        writer.write(file=file, root=root)
    with pytest.raises(OSError, match=r'File exists'):
        writer.write(file=file, root=root, mode='x')
    with pytest.raises(OSError, match=r'File exists'):
        writer.write(file=file, root=root, mode='w-')

    # invalid mode
    for m in ['r', 'b', 'w+b']:
        with pytest.raises(ValueError, match=r'Invalid mode'):
            writer.write(file=file, root=root, mode=m)

    # r+ is a valid mode, but the file must already exist
    with pytest.raises(OSError, match=r'File does not exist'):
        writer.write(file='does_not.exist', root=root, mode='r+')

    # by specifying the proper mode one can overwrite a file
    writer.write(file=file, root=root, mode='w')
    assert roots_equal(root, read(file))
    writer.write(file=file, root=root, mode='a')
    assert roots_equal(root, read(file))
    writer.write(file=file, root=root, mode='r+')
    assert roots_equal(root, read(file))
    os.remove(file)


@pytest.mark.skipif(h5py is None, reason='h5py not installed')
def test_numpy_unicode_dtype():
    writer = HDF5Writer()
    writer.add_metadata(wide_chars=np.array(['1', '-4e+99', 'True'], dtype='<U6'))
    writer.create_dataset('wide_chars', data=np.random.random(100).reshape(4, 25).astype('<U32'))

    file = tempfile.gettempdir() + '/msl-hdf5-writer-temp.h5'
    writer.save(file, mode='w')

    root = read(file)
    assert np.array_equal(root.metadata.wide_chars, writer.metadata.wide_chars)

    # the following array_equal assertion fails so we iterate over all elements instead
    # assert np.array_equal(root.wide_chars.astype('<U32'), writer.wide_chars)
    for a, b in zip(root.wide_chars.astype('<U32').flatten(), writer.wide_chars.flatten()):
        assert a == b

    os.remove(file)



import os
from shutil import rmtree
from tempfile import mkdtemp

import pytest

from msl.io.google_api import GDrive
from msl.io.google_api import GSheets

# all Google API tests require the necessary "token.json" file to be
# available for a specific Google user's account
try:
    drive = GDrive(account='demo', read_only=False)
    sheet = GSheets(account='demo', read_only=False)
except:
    drive = None
    sheet = None


TMP_DIR = ''


def setup_module():
    global TMP_DIR
    TMP_DIR = mkdtemp(prefix='msl-io-test-')


def teardown_module():
    rmtree(TMP_DIR)
    if sheet is not None:
        sheet.close()
    if drive is not None:
        drive.close()


@pytest.mark.skipif(drive is None, reason='No GDrive CI demo account token')
@pytest.mark.skipif(sheet is None, reason='No GSheets CI demo account token')
def test_drive_sheets():
    filename = os.path.basename(__file__)
    drives = drive.shared_drives()
    drive_id = next(iter(drives))
    assert drives[drive_id] == 'MSL - Joe'
    assert drive_id == '0AAkJczPd2UhCUk9PVA'
    spreadsheet_id = sheet.create('spreadsheet', sheet_names=['a'])
    spreadsheet_2_id = sheet.create('spreadsheet-2')
    assert drive.path(spreadsheet_id) == 'My Drive/spreadsheet'
    files_id = drive.create_folder('Files', parent_id=drive_id)
    copy_sid = drive.copy(spreadsheet_id, files_id)
    assert drive.path(spreadsheet_id) == 'My Drive/spreadsheet'
    assert drive.path(copy_sid) == 'Drive/Files/Copy of spreadsheet'
    sheet.write('hi', copy_sid, 'A1')
    assert sheet.values(copy_sid, 'A1') == [['hi']]
    assert drive.path(spreadsheet_2_id) == 'My Drive/spreadsheet-2'
    assert sheet.sheet_names(spreadsheet_2_id) == ('Sheet1',)
    sheet.copy('a', copy_sid, spreadsheet_2_id)
    assert sheet.sheet_names(spreadsheet_2_id) == ('Sheet1', 'Copy of a')
    assert sheet.values(copy_sid, cells='A1', sheet='a') == [['hi']]
    assert drive.path(files_id) == 'Drive/Files'
    assert drive.folder_id('Files', parent_id=drive_id) == files_id
    assert drive.folder_id('Drive/Files', parent_id=drive_id) == files_id
    assert drive.is_folder('Files', parent_id=drive_id) is True
    assert drive.is_file(filename, folder_id=files_id) is False
    f_id = drive.upload(__file__, files_id)
    assert drive.path(f_id) == 'Drive/Files/' + filename
    assert drive.is_file(filename, folder_id=files_id) is True
    assert drive.file_id(filename, folder_id=files_id) == f_id
    help_id = drive.create_folder('help', parent_id=files_id)
    drive.move(f_id, help_id)
    assert drive.path(f_id) == 'Drive/Files/help/' + filename
    assert drive.is_folder('Drive/Files/help', parent_id=drive_id) is True
    assert drive.is_folder('Files/help', parent_id=drive_id) is True
    assert drive.is_folder('help', parent_id=files_id) is True
    assert drive.is_file('Files/help/' + filename, folder_id=drive_id) is True
    assert drive.is_file('help/' + filename, folder_id=files_id) is True
    assert drive.is_file(filename, folder_id=help_id) is True
    assert drive.is_file('help/' + filename) is False
    assert not os.path.isfile(os.path.join(TMP_DIR, filename))
    drive.download(f_id, save_to=TMP_DIR)
    assert os.path.isfile(os.path.join(TMP_DIR, filename))
    assert not os.path.isfile(os.path.join(TMP_DIR, 'new.py'))
    drive.download(f_id, save_to=os.path.join(TMP_DIR, 'new.py'))
    assert os.path.isfile(os.path.join(TMP_DIR, 'new.py'))
    drive.move(f_id, 'root')
    assert drive.path(f_id) == 'My Drive/' + filename
    drive.move(f_id, files_id)
    assert drive.path(f_id) == 'Drive/Files/' + filename
    iid = drive.create_folder('Files Backup')
    copy_f_id = drive.copy(f_id, iid)
    assert drive.path(copy_f_id) == 'My Drive/Files Backup/' + filename
    drive.delete(files_id)
    drive.delete(iid)
    drive.delete(spreadsheet_id)
    drive.delete(spreadsheet_2_id)
    drive.empty_trash()
